from . import Project
from . import ComObjectWrapper

class AnalyticalCurve(ComObjectWrapper):
    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.AnalyticalCurve

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, analytical_curve_name: str):
        self.invoke_method('Name', analytical_curve_name)

    def set_curve(self, curve_name: str):
        self.invoke_method('Curve', curve_name)

    def set_law_x(self, law: str):
        self.invoke_method('LawX', law)

    def set_law_y(self, law: str):
        self.invoke_method('LawY', law)

    def set_law_z(self, law: str):
        self.invoke_method('LawZ', law)

    def set_parameter_range(self, min: float, max: float):
        self.invoke_method('ParameterRange', min, max)

    def create(self):
        self.invoke_method('Create')


from . import Shape, IVBAProvider

class Sphere(Shape):
    AXIS_X = 'x'
    AXIS_Y = 'y'
    AXIS_Z = 'z'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Sphere')

    def set_axis(self, axis: str):
        self.record_method('Axis', axis)

    def set_center_radius(self, radius: float):
        self.record_method('CenterRadius', radius)

    def set_top_radius(self, radius: float):
        self.record_method('TopRadius', radius)

    def set_bottom_radius(self, radius: float):
        self.record_method('BottomRadius', radius)

    def set_center(self, x: float, y: float, z: float):
        self.record_method('Center', x, y, z)

    def set_num_segments(self, num_segments: int):
        self.record_method('Segments', num_segments)

    def set_smooth_geometry(self):
        self.set_num_segments(0)


from . import Project
from . import ComObjectWrapper
from . import w32com
import numpy as np

class PlaneWave(ComObjectWrapper):
    POLARIZATION_LINEAR = 'Linear'
    POLARIZATION_CIRCULAR = 'Circular'
    POLARIZATION_ELLIPTICAL = 'Elliptical'

    CIRCULAR_DIRECTION_LEFT = 'Left'
    CIRCULAR_DIRECTION_RIGHT = 'Right'

    DIRECTION_X = 'x'
    DIRECTION_Y = 'y'
    DIRECTION_Z = 'z'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.PlaneWave

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def store(self):
        self.invoke_method('Store')

    def delete(self):
        self.invoke_method('Delete')

    def set_normal_vector(self, x: float, y: float, z: float):
        self.invoke_method('Normal', x, y, z)

    def set_e_vector(self, x: float, y: float, z: float):
        self.invoke_method('EVector', x, y, z)

    def set_polarization(self, polarization_type: str):
        self.invoke_method('Polarization', polarization_type)

    def set_reference_frequency(self, frequency: float):
        self.invoke_method('ReferenceFrequency', frequency)

    def set_phase_difference_deg(self, angle: float):
        self.invoke_method('PhaseDifference', angle)

    def set_phase_difference_rad(self, angle: float):
        self.set_phase_difference_deg(np.rad2deg(angle))

    def set_circular_direction(self, direction: str):
        self.invoke_method('CircularDirection', direction)

    def set_axial_ratio(self, ratio: float):
        self.invoke_method('AxialRatio', ratio)

    def set_use_custom_decoupling_plane(self, flag: bool = True):
        self.invoke_method('SetUserDecouplingPlane', flag)

    def set_custom_decoupling_plane(self, direction: str, position: float):
        self.invoke_method('DecouplingPlane', direction, position)

    def get_normal_vector(self) -> tuple[float, float, float]:
        x = w32com.create_ref_double()
        y = w32com.create_ref_double()
        z = w32com.create_ref_double()
        self.invoke_method('GetNormal', x, y, z)
        return (x.value, y.value, z.value)

    def get_e_vector(self) -> tuple[float, float, float]:
        x = w32com.create_ref_double()
        y = w32com.create_ref_double()
        z = w32com.create_ref_double()
        self.invoke_method('GetEVector', x, y, z)
        return (x.value, y.value, z.value)

    def get_polarization_type(self) -> str:
        polarization_type = w32com.create_ref_str()
        self.invoke_method('GetPolarizationType', polarization_type)
        return polarization_type.value

    def get_circular_direction(self) -> str:
        dir = w32com.create_ref_str()
        self.invoke_method('GetCircularDirection', dir)
        return dir.value

    def get_reference_frequency(self) -> float:
        frequency = w32com.create_ref_double()
        self.invoke_method('GetReferenceFrequency', frequency)
        return frequency.value

    def get_phase_difference_deg(self) -> float:
        angle = w32com.create_ref_double()
        self.invoke_method('GetPhaseDifference', angle)
        return angle.value

    def get_phase_difference_rad(self) -> float:
        return np.deg2rad(self.get_phase_difference_deg())

    def get_axial_ratio(self) -> float:
        ratio = w32com.create_ref_double()
        self.invoke_method('GetAxialRatio', ratio)
        return ratio.value



from typing import Optional

class FrequencyRangeList:
    def __init__(self) -> None:
        self.items = []

    def clear(self):
        self.items = []

    def add_range(self, min_f: float, max_f: float):
        self.items.append((min_f, max_f))

    def add_single_point(self, freq: float):
        self.items.append(freq)

    def add_range_to_infinity(self, min_f: float):
        self.items.append((min_f, None))

    def __str__(self):
        items = [
            FrequencyRangeList.__convert_freq_ranges_list_item_to_str(x)
            for x in self.items ]
        return ';'.join(items)

    def __convert_freq_range_to_str(min_f: float, max_f: Optional[float]):
        if max_f is None:
            return str(min_f)+', '
        else:
            return str(min_f)+', '+str(max_f)

    def __convert_freq_ranges_list_item_to_str(item):
        if isinstance(item, tuple):
            return FrequencyRangeList.__convert_freq_range_to_str(item[0], item[1])
        else:
            return str(item)


from . import IVBAProvider, VBAObjWrapper

class EigenmodeSolver(VBAObjWrapper):
    METHOD_TYPE_AKS = 'AKS'
    METHOD_TYPE_JDM = 'JDM'
    METHOD_TYPE_JDM_LOW_MEM = 'JDM (low memory)'
    METHOD_TYPE_AUTOMATIC = 'Automatic'
    METHOD_TYPE_CLASSICAL = 'Classical (Lossless)'
    METHOD_TYPE_GENERAL = 'General (Lossy)'

    METHOD_TYPE_MESH_HEX = 'Hex'
    METHOD_TYPE_MESH_TET = 'Tet'

    MESH_TYPE_HEXAHEDRAL = 'Hexahedral Mesh'
    MESH_TYPE_TETRAHEDRAL = 'Tetrahedral Mesh'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'EigenmodeSolver')

    def reset(self):
        self.record_method('Reset')

    def start(self):
        return self.query_method_bool('Start')

    def set_mesh_adaptation_hex(self, flag: bool = True):
        self.record_method('SetMeshAdaptationHex', flag)

    def set_mesh_adaptation_tet(self, flag: bool = True):
        self.record_method('SetMeshAdaptationTet', flag)

    def set_number_of_modes(self, number: int):
        self.record_method('SetNumberOfModes', number)

    def set_calculate_modes_in_frequency_range(self, flag: bool = True):
        self.record_method('SetModesInFrequencyRange', flag)

    def set_consider_static_modes(self, flag: bool = True):
        self.record_method('SetConsiderStaticModes', flag)

    def set_use_remote_calculation(self, flag: bool = True):
        self.record_method('SetRemoteCalculation', flag)

    def set_method_type(self, method_type: str, mesh: str):
        self.record_method('SetMethodType', method_type, mesh)

    def set_mesh_type(self, mesh_type: str):
        self.record_method('SetMeshType', mesh_type)

    def set_material_eval_frequency(self, freq: float):
        if freq is None:
            self.record_method('SetMaterialEvaluationFrequency', False, 0.0)
        else:
            self.record_method('SetMaterialEvaluationFrequency', True, freq)

    def set_frequency_target(self, freq: float):
        if freq is None:
            self.record_method('SetFrequencyTarget', False, 0.0)
        else:
            self.record_method('SetFrequencyTarget', True, freq)

    def set_lower_bound_for_q(self, q_min: float):
        if q_min is None:
            self.record_method('SetLowerBoundForQ', False, 0.0)
        else:
            self.record_method('SetLowerBoundForQ', True, q_min)

    def set_lower_bound_for_q(self, flag: bool = True):
        self.record_method('SetLowerBoundForQ', flag, '')

    def set_max_number_of_threads(self, number: int):
        self.record_method('SetMaxNumberOfThreads', number)

    def set_use_parallelization(self, flag: bool = True):
        self.record_method('SetUseParallelization', flag)

    def set_consider_losses_in_postprocessing_only(self, flag: bool = True):
        self.record_method('SetConsiderLossesInPostprocessingOnly', flag)

    def get_consider_losses_in_postprocessing_only(self) -> bool:
        return self.query_method_bool('GetConsiderLossesInPostprocessingOnly')

    def set_minimum_q(self, min_q: float):
        self.record_method('SetMinimumQ', min_q)

    def set_calculate_external_q_factor(self, flag: bool = True):
        self.record_method('SetCalculateExternalQFactor', flag)

    def set_q_external_accuracy(self, accuracy: float):
        self.record_method('SetQExternalAccuracy', accuracy)

    def set_order_tet(self, order: int):
        self.record_method('SetOrderTet', order)

    def set_store_results_in_cache(self, flag: bool = True):
        self.record_method('SetStoreResultsInCache', flag)

    def set_td_compatible_materials(self, flag: bool = True):
        self.record_method('SetTDCompatibleMaterials', flag)

    def set_calculate_thermal_losses(self, flag: bool = True):
        self.record_method('SetCalculateThermalLosses', flag)

    def set_accuracy(self, accuracy: float):
        self.record_method('SetAccuracy', accuracy)

    def get_number_of_modes_calculated(self) -> int:
        return self.query_method_int('GetNumberOfModesCalculated')

    def get_mode_frequency_hz(self, mode_number: int) -> float:
        return self.query_method_float('GetModeFrequencyInHz', mode_number)

    def get_mode_rel_residual_norm(self, mode_number: int) -> float:
        return self.query_method_float('GetModeRelResidualNorm', mode_number)

    def get_mode_q_factor(self, mode_number: int) -> float:
        return self.query_method_float('GetModeQFactor', mode_number)

    def get_mode_external_q_factor(self, mode_number: int) -> float:
        return self.query_method_float('GetModeExternalQFactor', mode_number)

    def get_loaded_frequency_hz(self, mode_number: int) -> float:
        return self.query_method_float('GetLoadedFrequencyInHz', mode_number)

    def get_number_of_sensitivity_design_parameters(self) -> int:
        return self.query_method_int('GetNumberOfSensitivityDesignParameters')

    # indexed from 0
    def get_sensitivity_design_parameter(self, index: int) -> str:
        return self.query_method_str('GetSensitivityDesignParameter', index+1)

    def get_frequency_sensitivity(self, param_name: str, mode_number: int) -> float:
        return self.query_method_float('GetFrequencySensitivity', param_name, mode_number)

    def get_q_factor_sensitivity(self, param_name: str, mode_number: int) -> float:
        return self.query_method_float('GetQFactorSensitivity', param_name, mode_number)

    def reset_force_calculation(self):
        self.record_method('ResetForceCalculation')

    def calculate_lorentz_force_for_mode(self, mode_index: int):
        self.record_method('CalculateLorentzForceForMode', mode_index)

    def calculate_lorentz_force_for_all_modes(self):
        self.record_method('CalculateLorentzForceForAllModes')

    def is_mode_selected_for_force_calculation(self, mode_index: int) -> bool:
        return self.query_method_bool('IsModeSelectedForForceCalculation', mode_index)

    def is_any_mode_selected_for_force_calculation(self) -> bool:
        return self.query_method_bool('IsAnyModeSelectedForForceCalculation')

    def start_force_calculation(self) -> bool:
        return self.query_method_bool('StartForceCalculation')


from . import Project
from . import ComObjectWrapper

class ResultTree(ComObjectWrapper):
    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.ResultTree

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def enable_update(self, flag: bool = True):
        self.invoke_method('EnableTreeUpdate', flag)

    def disable_update(self, flag: bool = True):
        self.invoke_method('EnableTreeUpdate', not flag)

    def get_first_child_name(self, parent_item: str) -> str:
        return self.invoke_method('GetFirstChildName', parent_item)

    def get_next_item_name(self, current_item: str) -> str:
        return self.invoke_method('GetNextItemName', current_item)

    def refresh_view(self):
        self.invoke_method('RefreshView')

    def has_item(self, item: str):
        return self.invoke_method('DoesTreeItemExist', item)


from . import VBATypeName
from typing import Union, Optional

def construct_function(function_name: str, *args: Union[str, int, float, bool]) -> str:
    args_as_strings = [f'"{arg}"' if isinstance(arg, str) else str(arg) for arg in args]
    args_str = ", ".join(args_as_strings)
    return f'{function_name}({args_str})'

def construct_method(
        object_name: str, function_name: str, *args: Union[str, int, float, bool]) -> str:
    args_as_strings = [f'"{arg}"' if isinstance(arg, str) else str(arg) for arg in args]
    args_str = ", ".join(args_as_strings)
    return f'{object_name}.{function_name}({args_str})'

def get_num_vars(*args: Union[str, int, float, bool, VBATypeName]):
    num_vars = 0
    for arg in args:
        if isinstance(arg, VBATypeName):
            num_vars += 1
    return num_vars

def generate_variables_lines(
        return_type: Optional[VBATypeName], *args: Union[str, int, float, bool, VBATypeName]):
    variables = ''
    if return_type is not None:
        variables += f'Dim r As {return_type.value}\n'
    arg_cntr = 0
    for arg in args:
        if isinstance(arg, VBATypeName):
            variables += f'Dim var{arg_cntr} As {arg.value}\n'
        arg_cntr += 1
    return variables

def generate_statement_line(
        object_name: Optional[str], function_name: str, return_type: Optional[VBATypeName],
        *args: Union[str, int, float, bool, VBATypeName]):
    statement = ''
    if return_type is not None:
        statement += 'r = '
    if object_name is not None:
        statement += f'{object_name}.'
    statement += f'{function_name}('
    arg_cntr = 0
    for arg in args:
        if arg_cntr != 0:
            statement += ','
        if isinstance(arg, VBATypeName):
            statement += f'var{arg_cntr}'
        elif isinstance(arg, str):
            statement += f'"{arg}"'
        else:
            statement += f'{arg}'
        arg_cntr += 1
    statement += ')'
    return statement

def generate_report_line(
        return_type: Optional[VBATypeName], *args: Union[str, int, float, bool, VBATypeName]):
    if get_num_vars(*args) == 0 and return_type is None:
        return ''
    report = 'ReportInformation('
    if return_type is not None:
        report += 'r'
        if get_num_vars(*args) > 0:
            report += ' & "," & '
    var_cntr = 0
    arg_cntr = 0
    for arg in args:
        if isinstance(arg, VBATypeName):
            if var_cntr != 0:
                report += ' & "," & '
            report += f'var{arg_cntr}'
            var_cntr += 1
        arg_cntr += 1
    report += ' & "")'
    return report

class IVBARuntime:
    def run_vba(self, vba_code: str) -> bool:
        return False

    def invoke_function(self, function_name: str, *args: Union[str, int, float, bool]):
        self.run_vba(construct_function(function_name, *args))

    def invoke_method(
            self, object_name: str, function_name: str, *args: Union[str, int, float, bool]):
        self.run_vba(construct_method(object_name, function_name, *args))

    def report_vba(self, vba_code: str) -> bool:
        return False

    def report_function(self, function_name: str, *args: Union[str, int, float, bool]):
        self.report_vba(construct_function(function_name, *args))

    def report_method(
            self, object_name: str, function_name: str, *args: Union[str, int, float, bool]):
        self.report_vba(construct_method(object_name, function_name, *args))

    def report_function_t(
            self, function_name: str, return_type: Optional[VBATypeName],
            *args: Union[str, int, float, bool, VBATypeName]):
        variables = generate_variables_lines(return_type, *args)
        statement = generate_statement_line(None, function_name, return_type, *args)
        report = generate_report_line(return_type, *args)
        self.run_vba(f'{variables}{statement}\n{report}')

    def report_method_t(
            self, object_name: str, function_name: str, return_type: Optional[VBATypeName],
            *args: Union[str, int, float, bool, VBATypeName]):
        variables = generate_variables_lines(return_type, *args)
        statement = generate_statement_line(object_name, function_name, return_type, *args)
        report = generate_report_line(return_type, *args)
        self.run_vba(f'{variables}{statement}\n{report}')



from __future__ import annotations
import re
from parse import parse
import numpy as np

class Farfield:
    def __init__(self) -> None:
        self.samples = None # [theta_index,phi_index,component_index:0=theta/1=phi]
        self.frequency = 0
        self.position = np.array([0., 0., 0.])
        self.z_axis = np.array([0., 0., 1.])
        self.x_axis = np.array([1., 0., 0.])
        self.radiated_power = 0
        self.accepted_power = 0
        self.stimulated_power = 0

    def __str__(self) -> str:
        templ = 'Farfield:\n'\
            '\tFrequency: {freq:.2e} Hz\n'\
            '\tPosition: {pos} m\n'\
            '\tZ axis: {z_axis} m\n'\
            '\tX axis: {x_axis} m\n'\
            '\tRadiated power: {r_pow:.2e} W\n'\
            '\tAccepted power: {a_pow:.2e} W\n'\
            '\tStimulated power: {s_pow:.2e} W\n'\
            '\t# Theta components: {num_theta}\n'\
            '\t# Phi components: {num_phi}'
        return templ.format(
            freq=self.frequency, pos=self.position, z_axis=self.z_axis, x_axis=self.x_axis,
            r_pow=self.radiated_power, a_pow=self.accepted_power, s_pow=self.stimulated_power,
            num_theta=self.get_num_theta_samples(), num_phi=self.get_num_phi_samples())

    def set_samples(self, samples: np.array):
        self.samples = samples

    def get_samples(self) -> np.array:
        return self.samples

    def set_theta_component(self, e_theta: np.array):
        self.samples[:,:,0] = e_theta

    def get_theta_component(self) -> np.array:
        return self.samples[:,:,0]

    def set_phi_component(self, e_phi: np.array):
        self.samples[:,:,1] = e_phi

    def get_phi_component(self) -> np.array:
        return self.samples[:,:,1]

    def get_sample_at(self, theta_index: int, phi_index: int) -> np.array:
        return self.samples[theta_index,phi_index,:]

    def init_samples_with_zeros(self, num_theta_samples: int, num_phi_samples: int):
        dim = (num_theta_samples, num_phi_samples, 2)
        self.samples = np.zeros(dim, dtype=np.complex128)

    def get_num_theta_samples(self) -> int:
        if self.samples is None:
            return 0
        return self.samples.shape[0]

    def get_theta_step_deg(self) -> float:
        self._ensure_existing_samples()
        return 180/(self.get_num_theta_samples() - 1)

    def get_theta_step_rad(self) -> float:
        self._ensure_existing_samples()
        return np.pi/(self.get_num_theta_samples() - 1)

    def get_theta_at_index_deg(self, index: int) -> float:
        return self.get_theta_step_deg()*index

    def get_theta_at_index_rad(self, index: int) -> float:
        return self.get_theta_step_rad()*index

    def get_index_of_theta_deg(self, angle_deg: float) -> int:
        return int(np.round(angle_deg/self.get_theta_step_deg()))

    def get_index_of_theta_rad(self, angle_rad: float) -> int:
        return int(np.round(angle_rad/self.get_theta_step_rad()))

    def get_theta_vec_deg(self) -> float:
        self._ensure_existing_samples()
        return np.arange(self.get_num_theta_samples())*self.get_theta_step_deg()

    def get_theta_vec_rad(self) -> float:
        self._ensure_existing_samples()
        return np.arange(self.get_num_theta_samples())*self.get_theta_step_rad()

    def get_num_phi_samples(self) -> int:
        if self.samples is None:
            return 0
        return self.samples.shape[1]

    def get_phi_step_deg(self) -> float:
        self._ensure_existing_samples()
        return 360/(self.get_num_phi_samples() - 1)

    def get_phi_step_rad(self) -> float:
        self._ensure_existing_samples()
        return 2*np.pi/(self.get_num_phi_samples() - 1)

    def get_phi_at_index_deg(self, index: int) -> float:
        return self.get_phi_step_deg()*index

    def get_phi_at_index_rad(self, index: int) -> float:
        return self.get_phi_step_rad()*index

    def get_index_of_phi_deg(self, angle_deg: float) -> int:
        return int(np.round(angle_deg/self.get_phi_step_deg()))

    def get_index_of_phi_rad(self, angle_rad: float) -> int:
        return int(np.round(angle_rad/self.get_phi_step_rad()))

    def get_phi_vec_deg(self) -> float:
        self._ensure_existing_samples()
        return np.arange(self.get_num_phi_samples())*self.get_phi_step_deg()

    def get_phi_vec_rad(self) -> float:
        self._ensure_existing_samples()
        return np.arange(self.get_num_phi_samples())*self.get_phi_step_rad()

    def set_frequency(self, frequency: float):
        self.frequency = frequency

    def get_frequency(self) -> float:
        return self.frequency

    def set_position(self, position: np.array):
        self.position = position

    def get_position(self) -> np.array:
        return self.position

    def set_z_axis(self, z_axis: np.array):
        self.z_axis = z_axis

    def get_z_axis(self) -> np.array:
        return self.z_axis

    def set_x_axis(self, x_axis: np.array):
        self.x_axis = x_axis

    def get_x_axis(self) -> np.array:
        return self.x_axis

    def set_radiated_power(self, power: float):
        self.radiated_power = power

    def get_radiated_power(self) -> float:
        return self.radiated_power

    def set_accepted_power(self, power: float):
        self.accepted_power = power

    def get_accepted_power(self) -> float:
        return self.accepted_power

    def set_stimulated_power(self, power: float):
        self.stimulated_power = power

    def get_stimulated_power(self) -> float:
        return self.stimulated_power

    def get_inner_product(ff1: Farfield, ff2: Farfield):
        Farfield._ensure_same_farfield_sizes(ff1, ff2)
        s1 = ff1.samples
        s2 = ff2.samples
        sin_theta_vec = np.array([np.sin(ff1.get_theta_vec_rad())]).T
        theta_grid = np.tile(sin_theta_vec, [1, ff1.get_num_phi_samples()])
        dS = 2*theta_grid*np.sin(ff1.get_theta_step_rad()/2)*ff1.get_phi_step_rad()
        integrand_theta = np.conj(s1[0:-2,0:-2,0]) * s2[0:-2,0:-2,0] * dS[0:-2,0:-2]
        integrand_phi = np.conj(s1[0:-2,0:-2,1]) * s2[0:-2,0:-2,1] * dS[0:-2,0:-2]
        integral_theta = np.sum(np.sum(integrand_theta))
        integral_phi = np.sum(np.sum(integrand_phi))
        return (integral_theta+integral_phi, integral_theta, integral_phi)

    def _ensure_existing_samples(self):
        if self.samples is None:
            raise Exception('Farfield has no samples!')

    def _ensure_same_farfield_sizes(ff1, ff2):
        if ff1.samples.shape != ff2.samples.shape:
            raise Exception('Failed to compute dot product: Farfields have different sizes')

class Parser:
    def __init__(self) -> None:
        self.reset()

    def reset(self):
        self.lines = None
        self.current_line = None
        self.version = None
        self.data_type = None
        self.num_frequencies = None
        self.position = None
        self.z_axis = None
        self.x_axis = None
        self.radiated_powers = []
        self.accepted_powers = []
        self.stimulated_powers = []
        self.frequencies = []
        self.farfields = []
        self.num_phi_samples = None
        self.num_theta_samples = None

    def parse_string(self, ffs_source_str: str):
        self.reset()
        self.lines = ffs_source_str.splitlines(keepends=False)
        self._process_lines()

    def get_farfields(self):
        farfields = []
        for n in range(len(self.frequencies)):
            farfields.append(self._get_nth_farfield(n))
        return farfields

    def _get_nth_farfield(self, n):
        farfield = Farfield()
        farfield.set_samples(self.farfields[n])
        farfield.set_frequency(self.frequencies[n])
        farfield.set_position(self.position)
        farfield.set_z_axis(self.z_axis)
        farfield.set_x_axis(self.x_axis)
        farfield.set_radiated_power(self.radiated_powers[n])
        farfield.set_accepted_power(self.accepted_powers[n])
        farfield.set_stimulated_power(self.stimulated_powers[n])
        return farfield

    def _process_lines(self):
        self.current_line = 0
        while self._has_lines_to_process():
            self._process_next_line()

    def _has_lines_to_process(self):
        return self.current_line < len(self.lines)

    def _get_line_rel(self, relative_line_index):
        return self.lines[self.current_line+relative_line_index]

    def _jump_to_line_rel(self, relative_line_index):
        self.current_line += relative_line_index

    def _process_next_line(self):
        line = self.lines[self.current_line]
        self._process_line(line)

    def _process_line(self, line):
        if 'Version' in line:
            self._process_version_line()
        elif 'Data Type' in line:
            self._process_data_type_line()
        elif '#Frequencies' in line:
            self._process_num_frequencies_line()
        elif line.find('Position') >= 0:
            self._process_position_line()
        elif 'zAxis' in line:
            self._process_z_axis_line()
        elif 'xAxis' in line:
            self._process_x_axis_line()
        elif 'Radiated/Accepted/Stimulated Power , Frequency' in line:
            self._process_powers_and_frequencies_line()
        elif '#phi' in line and '#theta' in line:
            self._process_num_phi_theta_samples_line()
        elif 'Phi, Theta, Re(E_Theta), Im(E_Theta), Re(E_Phi), Im(E_Phi)' in line:
            self._process_farfield_samples_line()
        else:
            self.current_line += 1

    def _process_version_line(self):
        version = self._get_line_rel(1).strip()
        Parser._ensure_supported_version(version)
        self.version = version
        self._jump_to_line_rel(2)
        # print('version:', version)

    def _process_data_type_line(self):
        self.data_type = self._get_line_rel(1).strip()
        self._jump_to_line_rel(2)
        # print('data type:', self.data_type)

    def _process_num_frequencies_line(self):
        self.num_frequencies = int(self._get_line_rel(1).strip())
        self._jump_to_line_rel(2)
        # print('num frequencies:', self.num_frequencies)

    def _process_position_line(self):
        self.position = Parser._parse_vector_line(self._get_line_rel(1), 3)
        self._jump_to_line_rel(2)
        # print('position:', self.position)

    def _process_z_axis_line(self):
        self.z_axis = Parser._parse_vector_line(self._get_line_rel(1), 3)
        # print('zAxis:', self.z_axis)
        self._jump_to_line_rel(2)

    def _process_x_axis_line(self):
        self.x_axis = Parser._parse_vector_line(self._get_line_rel(1), 3)
        # print('xAxis:', self.x_axis)
        self._jump_to_line_rel(2)

    def _process_powers_and_frequencies_line(self):
        self._jump_to_line_rel(1)
        self._process_all_powers_and_frequencies_lines()

    def _process_all_powers_and_frequencies_lines(self):
        for _ in range(self.num_frequencies):
            self._process_next_powers_and_frequency_lines()

    def _process_next_powers_and_frequency_lines(self):
        self.radiated_powers.append(float(self._get_line_rel(0).strip()))
        self.accepted_powers.append(float(self._get_line_rel(1).strip()))
        self.stimulated_powers.append(float(self._get_line_rel(2).strip()))
        self.frequencies.append(float(self._get_line_rel(3).strip()))
        self._jump_to_line_rel(5)
        # print('R/A/S/F', self.radiated_powers[-1], self.accepted_powers[-1], self.stimulated_powers[-1], self.frequencies[-1])

    def _process_num_phi_theta_samples_line(self):
        values = Parser._parse_vector_line(self._get_line_rel(1), 2)
        self.num_phi_samples = int(values[0])
        self.num_theta_samples = int(values[1])
        self._reserve_new_farfield_tensor(self.num_phi_samples, self.num_theta_samples)
        self._jump_to_line_rel(2)
        # print('#phi:', self.num_phi_samples)
        # print('#theta', self.num_theta_samples)

    def _reserve_new_farfield_tensor(self, num_phi_samples, num_theta_samples):
        dim = (num_theta_samples, num_phi_samples, 2)
        self.farfields.append(np.zeros(dim, dtype=np.complex128))

    def _process_farfield_samples_line(self):
        self._jump_to_line_rel(1)
        self._process_all_farfield_sample_data_lines()

    def _process_all_farfield_sample_data_lines(self):
        farfield_tensor = self.farfields[-1]
        for phi_i in range(self.num_phi_samples):
            for theta_i in range(self.num_theta_samples):
                self._process_next_farfield_sample_data_line(farfield_tensor, phi_i, theta_i)

    def _process_next_farfield_sample_data_line(self, farfield_tensor, phi_i, theta_i):
        values = Parser._parse_vector_line(self._get_line_rel(0), 6)
        self._jump_to_line_rel(1)
        # print('|', len(self.farfields), phi_i, theta_i, '| farfield data:', values)
        farfield_tensor[theta_i, phi_i, 0] = values[2] + 1j*values[3]
        farfield_tensor[theta_i, phi_i, 1] = values[4] + 1j*values[5]

    def _ensure_supported_version(version: str):
        if not Parser._is_supported_version(version):
            raise Exception('Unsupported ffs version: ' + version)

    def _is_supported_version(version: str):
        return version == '3.0'

    def _parse_vector_line(line, num_components: int):
        line = Parser._remove_duplicate_spaces(line.strip())
        vec = parse(('{} '*num_components).rstrip(), line).fixed
        Parser._ensure_vector_line_num_components(vec, num_components)
        return np.array([float(x) for x in vec])

    def _ensure_vector_line_num_components(vec, expected_num_components: int):
        if len(vec) != expected_num_components:
            raise Exception(
                'Failed to parse vector line: Invalid number of components - expected '
                + str(expected_num_components) + ', got ' + str(len(vec)))

    def _remove_duplicate_spaces(txt: str):
        return re.sub(' +', ' ', txt)

class Dumper:
    def __init__(self) -> None:
        self.reset()

    def reset(self):
        self.farfields = []
        self.version = '3.0'
        self.data_type = 'Farfield'

    def set_version(self, version: str):
        Dumper._ensure_supported_version(version)
        self.version = version

    def set_farfields(self, farfields: list):
        self.farfields = farfields

    def add_farfield(self, ff: Farfield):
        self.farfields.append(ff)

    def clear_farfields(self):
        self.farfields.clear()

    def dump_to_string(self) -> str:
        res = self._dump_header()
        res += self._dump_version()
        res += self._dump_data_type()
        res += self._dump_num_freq()
        res += self._dump_position()
        res += self._dump_z_axis()
        res += self._dump_x_axis()
        res += self._dump_powers_and_frequencies()
        res += self._dump_samples()
        return res

    def _dump_header(self) -> str:
        return '// CST Farfield Source File\n \n'

    def _dump_version(self) -> str:
        return '// Version:\n{ver} \n\n'.format(ver=self.version)

    def _dump_data_type(self) -> str:
        return '// Data Type\n{dt} \n\n'.format(dt=self.data_type)

    def _dump_num_freq(self) -> str:
        return '// #Frequencies\n{nf} \n\n'.format(nf=len(self.farfields))

    def _dump_position(self) -> str:
        pos = self.farfields[0].get_position()
        return '// Position\n{vec}\n\n'.format(vec=Dumper._dump_vec(pos))

    def _dump_z_axis(self) -> str:
        pos = self.farfields[0].get_z_axis()
        return '// zAxis\n{vec}\n\n'.format(vec=Dumper._dump_vec(pos))

    def _dump_x_axis(self) -> str:
        pos = self.farfields[0].get_x_axis()
        return '// xAxis\n{vec}\n\n'.format(vec=Dumper._dump_vec(pos))

    def _dump_powers_and_frequencies(self) -> str:
        templ = '// Radiated/Accepted/Stimulated Power , Frequency \n'+\
            ('{}'*len(self.farfields))+'\n'
        values = map(lambda ff: Dumper._dump_powers_and_frequency(ff), self.farfields)
        return templ.format(*values)

    def _dump_samples(self):
        res = ''
        for ff in self.farfields:
            res += Dumper._dump_ff_num_samples(ff)+Dumper._dump_ff_samples(ff)
        return res

    def _ensure_supported_version(version: str):
        if not Dumper._is_supported_version(version):
            raise Exception('Unsupported ffs version: ' + version)

    def _is_supported_version(version: str):
        return version == '3.0'

    def _dump_vec(vec):
        return ('{:e} '*len(vec)).format(*vec)

    def _dump_powers_and_frequency(ff: Farfield) -> str:
        return '{:e} \n{:e} \n{:e} \n{:e} \n\n'.format(
            ff.get_radiated_power(), ff.get_accepted_power(),
            ff.get_stimulated_power(), ff.get_frequency())

    def _dump_ff_num_samples(ff: Farfield) -> str:
        templ = '// >> Total #phi samples, total #theta samples\n{n_phi} {n_theta}\n\n'
        return templ.format(n_phi=ff.get_num_phi_samples(), n_theta=ff.get_num_theta_samples())

    def _dump_ff_samples(ff: Farfield) -> str:
        templ = '// >> Phi, Theta, Re(E_Theta), Im(E_Theta), Re(E_Phi), Im(E_Phi): \n{samples}\n'
        return templ.format(samples=Dumper._dump_ff_sample_vecs(ff))

    def _dump_ff_sample_vecs(ff: Farfield) -> str:
        res = ''
        for phi_index in range(ff.get_num_phi_samples()):
            for theta_index in range(ff.get_num_theta_samples()):
                res += Dumper._dump_sample_at(ff, phi_index, theta_index)
        return res

    def _dump_sample_at(ff: Farfield, phi_index, theta_index):
        phi = ff.get_phi_at_index_deg(phi_index)
        theta = ff.get_theta_at_index_deg(theta_index)
        field_vec = ff.get_sample_at(theta_index, phi_index)
        e_theta = field_vec[0]
        e_phi = field_vec[1]
        templ = '{:>8.3f}  {:>8.3f} {:>16.8e} {:>16.8e} {:>15.8e} {:>16.8e}\n'
        return templ.format(
            phi, theta, np.real(e_theta), np.imag(e_theta),
            np.real(e_phi), np.imag(e_phi))


from . import ResultItem

class ResultModule:
    def __init__(self, native_obj) -> None:
        self.native_obj = native_obj

    def get_all_run_ids(self, max_mesh_passes_only: bool = True) -> list[int]:
        return self.native_obj.get_all_run_ids(max_mesh_passes_only)

    def get_parameter_combination(self, run_id: int) -> dict:
        return self.native_obj.get_parameter_combination(run_id)

    def get_result_item(
            self, tree_path: str, run_id: int = 0, load_impedances: bool = True) -> ResultItem:
        return ResultItem(self.native_obj.get_result_item(tree_path, run_id, load_impedances))

    def get_run_ids(self, tree_path: str, skip_non_parametric: bool = False) -> list[int]:
        return self.native_obj.get_run_ids(tree_path, skip_non_parametric)

    def get_tree_items(self, item_filter: str = '0D/1D') -> list[str]:
        return self.native_obj.get_tree_items(item_filter)


from . import IVBAProvider, VBAObjWrapper

class MeshSettings(VBAObjWrapper):
    # This object is completely undocumented in CST Studio Suite Help, found in the history list

    MESH_TYPE_HEX = 'Hex'
    MESH_TYPE_HEX_TLM = 'HexTLM'
    MESH_TYPE_TET = 'Tet'
    MESH_TYPE_UNSTR = 'Unstr'
    MESH_TYPE_ALL = 'All'

    def __init__(self, vbap: IVBAProvider) -> None:
        super.__init__(vbap, 'MeshSettings')

    def set_mesh_type(self, mesh_type: str):
        self.record_method('SetMeshType', mesh_type)

    def set_version(self, version: int):
        self.record_method('Set', 'Version', version)

    def set_number_of_steps_per_wave_near(self, number: float):
        self.record_method('Set', 'StepsPerWaveNear', number)

    def set_number_of_steps_per_wave_far(self, number: float):
        self.record_method('Set', 'StepsPerWaveFar', number)

    def set_wavelength_refinement_same_as_near(self, flag: bool = True):
        self.record_method('Set', 'WavelengthRefinementSameAsNear', 1 if flag else 0)

    def set_number_of_steps_per_box_near(self, number: float):
        self.record_method('Set', 'StepsPerBoxNear', number)

    def set_number_of_steps_per_box_far(self, number: float):
        self.record_method('Set', 'StepsPerBoxFar', number)

    def set_max_number_of_steps_near(self, number: float):
        self.record_method('Set', 'MaxStepNear', number)

    def set_max_number_of_steps_far(self, number: float):
        self.record_method('Set', 'MaxStepFar', number)

    def set_model_box_descr_near(self, descr: str):
        self.record_method('Set', 'ModelBoxDescrNear', descr)

    def set_model_box_descr_far(self, descr: str):
        self.record_method('Set', 'ModelBoxDescrFar', descr)

    def set_use_max_step_absolute(self, flag: bool = True):
        self.record_method('Set', 'UseMaxStepAbsolute', 1 if flag else 0)

    def set_geometry_refinement_same_as_near(self, flag: bool = True):
        self.record_method('Set', 'WavelengthRefinementSameAsNear', 1 if flag else 0)

    def set_use_ratio_limit_geometry(self, flag: bool = True):
        self.record_method('Set', 'UseRatioLimitGeometry', 1 if flag else 0)

    def set_ratio_limit_geometry(self, limit: float):
        self.record_method('Set', 'RatioLimitGeometry', limit)

    def set_min_step_geometry_x(self, step: float):
        self.record_method('Set', 'MinStepGeometryX', step)

    def set_min_step_geometry_y(self, step: float):
        self.record_method('Set', 'MinStepGeometryY', step)

    def set_min_step_geometry_z(self, step: float):
        self.record_method('Set', 'MinStepGeometryZ', step)

    def set_min_step_geometry_xyz(self, x: float, y: float, z: float):
        self.set_min_step_geometry_x(x)
        self.set_min_step_geometry_y(y)
        self.set_min_step_geometry_z(z)

    def set_use_same_min_step_geometry(self, flag: bool = True):
        self.record_method('Set', 'UseSameMinStepGeometryXYZ', 1 if flag else 0)

    def set_plane_merge_version(self, version: str):
        self.record_method('Set', 'PlaneMergeVersion', version)

    def set_face_refinement_on(self, flag: bool = True):
        self.record_method('Set', 'FaceRefinementOn', 1 if flag else 0)

    def set_face_refinement_policy(self, value: int):
        self.record_method('Set', 'FaceRefinementPolicy', value)

    def set_face_refinement_ratio(self, value: float):
        self.record_method('Set', 'FaceRefinementRatio', value)

    def set_face_refinement_step(self, value: float):
        self.record_method('Set', 'FaceRefinementStep', value)

    def set_face_refinement_number_of_steps(self, number: int):
        self.record_method('Set', 'FaceRefinementNSteps', number)

    def set_face_refinement_number_of_buffer_lines(self, number: int):
        self.record_method('Set', 'FaceRefinementBufferLines', number)

    def set_ellipse_refinement_on(self, flag: bool = True):
        self.record_method('Set', 'EllipseRefinementOn', 1 if flag else 0)

    def set_ellipse_refinement_policy(self, value: int):
        self.record_method('Set', 'EllipseRefinementPolicy', value)

    def set_ellipse_refinement_ratio(self, value: float):
        self.record_method('Set', 'EllipseRefinementRatio', value)

    def set_ellipse_refinement_step(self, value: float):
        self.record_method('Set', 'EllipseRefinementStep', value)

    def set_ellipse_refinement_number_of_steps(self, number: int):
        self.record_method('Set', 'EllipseRefinementNSteps', number)

    def set_edge_refinement_on(self, flag: bool = True):
        self.record_method('Set', 'EdgeRefinementOn', 1 if flag else 0)

    def set_edge_refinement_policy(self, value: int):
        self.record_method('Set', 'EdgeRefinementPolicy', value)

    def set_edge_refinement_ratio(self, value: float):
        self.record_method('Set', 'EdgeRefinementRatio', value)

    def set_edge_refinement_step(self, value: float):
        self.record_method('Set', 'EdgeRefinementStep', value)

    def set_edge_refinement_number_of_buffer_lines(self, number: int):
        self.record_method('Set', 'EdgeRefinementBufferLines', number)

    def set_refine_edge_material_global(self, flag: bool = True):
        self.record_method('Set', 'RefineEdgeMaterialGlobal', 1 if flag else 0)

    def set_refine_axial_edge_global(self, flag: bool = True):
        self.record_method('Set', 'RefineAxialEdgeGlobal', 1 if flag else 0)

    def set_number_of_buffer_lines_near(self, number: int):
        self.record_method('Set', 'BufferLinesNear', number)

    def set_use_dielectrics(self, flag: bool = True):
        self.record_method('Set', 'UseDielectrics', 1 if flag else 0)

    def set_equilibrate_on(self, flag: bool = True):
        self.record_method('Set', 'EquilibrateOn', 1 if flag else 0)

    def set_equilibrate(self, value: float):
        self.record_method('Set', 'Equilibrate', value)

    def set_ignore_thin_panel_material(self, flag: bool = True):
        self.record_method('Set', 'IgnoreThinPanelMaterial', 1 if flag else 0)

    def set_snap_to_axial_edges(self, flag: bool = True):
        self.record_method('Set', 'SnapToAxialEdges', 1 if flag else 0)

    def set_snap_to_planes(self, flag: bool = True):
        self.record_method('Set', 'SnapToPlanes', 1 if flag else 0)

    def set_snap_to_spheres(self, flag: bool = True):
        self.record_method('Set', 'SnapToSpheres', 1 if flag else 0)

    def set_snap_to_ellipses(self, flag: bool = True):
        self.record_method('Set', 'SnapToEllipses', 1 if flag else 0)

    def set_snap_to_cylinders(self, flag: bool = True):
        self.record_method('Set', 'SnapToCylinders', 1 if flag else 0)

    def set_snap_to_cylinder_centers(self, flag: bool = True):
        self.record_method('Set', 'SnapToCylinderCenters', 1 if flag else 0)

    def set_snap_to_ellipse_centers(self, flag: bool = True):
        self.record_method('Set', 'SnapToEllipseCenters', 1 if flag else 0)

    def set_phase_error_near(self, error: float):
        self.record_method('Set', 'PhaseErrorNear', error)

    def set_phase_error_far(self, error: float):
        self.record_method('Set', 'PhaseErrorFar', error)

    def set_cells_per_wavelength_policy(self, policy: str):
        self.record_method('Set', 'CellsPerWavelengthPolicy', policy)

    def set_min_step(self, value: float):
        self.record_method('Set', 'MinStep', value)

    def set_method(self, method: str):
        self.record_method('Set', 'Method', method)

    def set_curvature_order(self, order: int):
        self.record_method('Set', 'CurvatureOrder', order)

    def set_curvature_order_policy(self, policy: str):
        self.record_method('Set', 'CurvatureOrderPolicy', policy)

    def set_curv_refinement_control(self, value: str):
        self.record_method('Set', 'CurvRefinementControl', value)

    def set_normal_tolerance(self, value: float):
        self.record_method('Set', 'NormalTolerance', value)

    def set_srf_mesh_gradation(self, value: float):
        self.record_method('Set', 'SrfMeshGradation', value)

    def set_srf_mesh_optimization(self, flag: bool = True):
        self.record_method('Set', 'SrfMeshOptimization', 1 if flag else 0)

    def set_use_materials(self, flag: bool = True):
        self.record_method('Set', 'UseMaterials', 1 if flag else 0)

    def set_move_mesh(self, flag: bool = True):
        self.record_method('Set', 'MoveMesh', 1 if flag else 0)

    def set_automatic_edge_refinement(self, flag: bool = True):
        self.record_method('Set', 'AutomaticEdgeRefinement', 1 if flag else 0)

    def set_use_aniso_curve_refinement(self, flag: bool = True):
        self.record_method('Set', 'UseAnisoCurveRefinement', 1 if flag else 0)

    def set_use_same_srf_and_vol_mesh_gradation(self, flag: bool = True):
        self.record_method('Set', 'UseSameSrfAndVolMeshGradation', 1 if flag else 0)

    def set_vol_mesh_gradation(self, value: float):
        self.record_method('Set', 'VolMeshGradation', value)

    def set_vol_mesh_optimization(self, flag: bool = True):
        self.record_method('Set', 'VolMeshOptimization', 1 if flag else 0)

    def set_small_feature_size(self, value: float):
        self.record_method('Set', 'SmallFeatureSize', value)

    def set_coincidence_tolerance(self, value: float):
        self.record_method('Set', 'CoincidenceTolerance', value)

    def set_self_intersection_check(self, flag: bool = True):
        self.record_method('Set', 'SelfIntersectionCheck', 1 if flag else 0)

    def set_optimize_for_planar_structures(self, flag: bool = True):
        self.record_method('Set', 'OptimizeForPlanarStructures', 1 if flag else 0)


from . import IVBAProvider, VBAObjWrapper

class Group(VBAObjWrapper):
    TYPE_NORMAL = 'normal'
    TYPE_MESH = 'mesh'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Group')

    def create(self, name: str, group_type: str = TYPE_NORMAL):
        self.record_method('Add', name, group_type)

    def delete(self, name: str):
        self.record_method('Delete', name)

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def add_item(self, item_name: str, group_name: str):
        self.record_method('AddItem', item_name, group_name)

    def remove_item(self, item_name: str, group_name: str):
        self.record_method('RemoveItem', item_name, group_name)

    def create_folder(self, name: str):
        self.record_method('NewFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.record_method('RenameFolder', old_name, new_name)

    def delete_folder(self, name: str):
        self.record_method('DeleteFolder', name)

    def reset(self):
        self.record_method('Reset')


from . import Project
from . import ComObjectWrapper

class FarfieldSource(ComObjectWrapper):
    ALIGNMENT_CUSTOM = 'user'
    ALIGNMENT_CURRENT_WCS = 'currentwcs'
    ALIGNMENT_SOURCE_FILE = 'sourcefile'

    CALC_MODE_AUTOMATIC = 'automatic'
    CALC_MODE_CUSTOM = 'user defined'
    CALC_MODE_AUTOTRUNCATION = 'autotruncation'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.FarfieldSource

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_id(self, id: int):
        self.invoke_method('Id', id)

    def set_position(self, x: float, y: float, z: float):
        self.invoke_method('SetPosition', x, y, z)

    def set_phi0_vec(self, x: float, y: float, z: float):
        self.invoke_method('SetPhi0XYZ', x, y, z)

    def set_theta0_vec(self, x: float, y: float, z: float):
        self.invoke_method('SetTheta0XYZ', x, y, z)

    def import_from_file(self, file_path: str):
        self.invoke_method('Import', file_path)

    def set_use_copy_only(self, flag: bool = True):
        self.invoke_method('UseCopyOnly', flag)

    def set_use_multipole_ffs(self, flag: bool = True):
        self.invoke_method('UseMultipoleFFS', flag)

    def set_alignment_type(self, alignment_type: str):
        self.invoke_method('SetAlignmentType', alignment_type)

    def set_multipole_degree(self, degree: int):
        self.invoke_method('SetMultipoleDegree', degree)

    def set_multipole_calc_mode(self, mode: str):
        self.invoke_method('SetMultipoleCalcMode', mode)

    def store(self):
        self.invoke_method('Store')

    def delete(self):
        self.invoke_method('Delete')

    def delete_all(self):
        self.invoke_method('DeleteAll')

    def get_next_id(self) -> int:
        return self.invoke_method('GetNextId')


from . import IVBAProvider, VBAObjWrapper
import numpy as np

class AsymptoticSolver(VBAObjWrapper):
    SOLVER_TYPE_SBR = 'SBR'
    SOLVER_TYPE_SBR_RAYTUBES = 'SBR_RAYTUBES'

    SOLVER_MODE_MONOSTATIC_SCATTERING = 'MONOSTATIC_SCATTERING'
    SOLVER_MODE_BISTATIC_SCATTERING = 'BISTATIC_SCATTERING'
    SOLVER_MODE_FIELD_SOURCES = 'FIELD_SOURCES'
    SOLVER_MODE_RANGE_PROFILES = 'RANGE_PROFILES'

    ACCURACY_LOW = 'LOW'
    ACCURACY_MEDIUM = 'MEDIUM'
    ACCURACY_HIGH = 'HIGH'
    ACCURACY_CUSTOM = 'CUSTOM'

    RANGE_PROFILES_WINDOW_RECTANGULAR = 'RECTANGULAR'
    RANGE_PROFILES_WINDOW_HANNING = 'HANNING'
    RANGE_PROFILES_WINDOW_HAMMING = 'HAMMING'
    RANGE_PROFILES_WINDOW_BLACKMAN = 'BLACKMAN'

    RANGE_PROFILES_MODE_RANGE_EXTEND = 'RANGE_EXTEND'
    RANGE_PROFILES_MODE_BANDWIDTH = 'BANDWIDTH'

    ANGLE_SWEEP_POINT = 'POINT'
    ANGLE_SWEEP_THETA = 'THETA'
    ANGLE_SWEEP_PHI = 'PHI'
    ANGLE_SWEEP_BOTH = 'BOTH'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'AsymptoticSolver')

    def set_solver_type(self, solver_type: str):
        self.record_method('SetSolverType', solver_type)

    def set_solver_mode(self, solver_mode: str):
        self.record_method('SetSolverMode', solver_mode)

    def set_accuracy_level(self, accuracy_level: str):
        self.record_method('SetAccuracyLevel', accuracy_level)

    def set_solver_store_results_as_tables_only(self, flag: bool = True):
        self.record_method('SetSolverStoreResultsAsTablesOnly', flag)

    def set_calculate_rcs_map_for_1d_sweeps(self, flag: bool = True):
        self.record_method('CalculateRCSMapFor1DSweeps', flag)

    def set_calculate_monitors(self, flag: bool = True):
        self.record_method('Set', 'CalculateMonitors', flag)

    def reset_polarizations(self):
        self.record_method('ResetPolarizations')

    def add_horizontal_polarization(self, value: float):
        self.record_method('AddHorizontalPolarization', value)

    def add_vertical_polarization(self, value: float):
        self.record_method('AddVerticalPolarization', value)

    def add_lhc_polarization(self, value: float):
        self.record_method('AddLHCPolarization', value)

    def add_rhc_polarization(self, value: float):
        self.record_method('AddRHCPolarization', value)

    def add_custom_polarization(self, theta: complex, phi: complex):
        self.record_method('AddCustomPolarization', theta.real, theta.imag, phi.real, phi.imag)

    def set_solver_max_number_of_reflections(self, number: int):
        self.record_method('SetSolverMaximumNumberOfReflections', number)

    def set_solver_range_profiles_center_frequency(self, freq: float):
        self.record_method('SetSolverRangeProfilesCenterFrequency', freq)

    def set_solver_range_profiles_automatic(self, flag: bool = True):
        self.record_method('SetSolverRangeProfilesAutomatic', flag)

    def set_solver_range_profiles_number_of_samples(self, number: int):
        self.record_method('SetSolverRangeProfilesNumberOfSamples', number)

    def set_solver_range_profiles_window_function(self, window: str):
        self.record_method('SetSolverRangeProfilesWindowFunction', window)

    def set_solver_range_profiles_spec_mode(self, mode: str):
        self.record_method('SetSolverRangeProfilesSpecMode', mode)

    def set_solver_range_profiles_range_extend(self, ext: float):
        self.record_method('SetSolverRangeProfilesRangeExtend', ext)

    def set_solver_range_profiles_bandwidth(self, bw: float):
        self.record_method('SetSolverRangeProfilesBandwidth', bw)

    def reset_frequency_list(self):
        self.record_method('ResetFrequencyList')

    def add_frequency_sweep(self, f_min: float, f_max: float, f_step: float):
        self.record_method('AddFrequencySweep', f_min, f_max, f_step)

    def reset_excitation_angle_list(self):
        self.record_method('ResetExcitationAngleList')

    def add_excitation_angle_sweep_deg(
            self, angle_sweep_type: str, theta_min: float, theta_max: float, theta_step: float,
            phi_min: float, phi_max: float, phi_step: float):
        self.record_method(
            'AddExcitationAngleSweep', angle_sweep_type, theta_min, theta_max, theta_step,
            phi_min, phi_max, phi_step)

    def add_excitation_angle_sweep_rad(
            self, angle_sweep_type: str, theta_min: float, theta_max: float, theta_step: float,
            phi_min: float, phi_max: float, phi_step: float):
        self.add_excitation_angle_sweep_deg(
            angle_sweep_type,
            np.rad2deg(theta_min), np.rad2deg(theta_max), np.rad2deg(theta_step),
            np.rad2deg(phi_min), np.rad2deg(phi_max), np.rad2deg(phi_step))

    def add_excitation_angle_sweep_with_rays_deg(
            self, angle_sweep_type: str, theta_min: float, theta_max: float, theta_step: float,
            phi_min: float, phi_max: float, phi_step: float):
        self.record_method(
            'AddExcitationAngleSweepWithRays', angle_sweep_type,
            theta_min, theta_max, theta_step, phi_min, phi_max, phi_step)

    def add_excitation_angle_sweep_with_rays_rad(
            self, angle_sweep_type: str, theta_min: float, theta_max: float, theta_step: float,
            phi_min: float, phi_max: float, phi_step: float):
        self.add_excitation_angle_sweep_with_rays_deg(
            angle_sweep_type,
            np.rad2deg(theta_min), np.rad2deg(theta_max), np.rad2deg(theta_step),
            np.rad2deg(phi_min), np.rad2deg(phi_max), np.rad2deg(phi_step))

    def reset_field_sources(self):
        self.record_method('ResetFieldSources')

    def set_field_source_active(self, field_source_name: str, flag: bool = True):
        self.record_method('SetFieldSourceActive', field_source_name, flag)

    def set_field_source_phasor(self, field_source_name: str, amplitude: float, phase: float):
        self.record_method('SetFieldSourcePhasor', field_source_name, amplitude, phase)

    def set_field_source_store_rays(self, field_source_name: str, flag: bool = True):
        self.record_method('SetFieldSourceRays', field_source_name, flag)

    def set_simultaneous_field_source_excitation(self, flag: bool = True):
        self.record_method('SimultaneousFieldSourceExcitation', flag)

    def set_calculate_s_params(self, flag: bool = True):
        self.record_method('Set', 'CalculateSParameters', flag)

    def reset_observation_angle_list(self):
        self.record_method('ResetObservationAngleList')


from . import Shape, IVBAProvider

class AnalyticalFace(Shape):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'AnalyticalFace')

    def set_law_x(self, expr: str):
        self.record_method('LawX', expr)

    def set_law_y(self, expr: str):
        self.record_method('LawY', expr)

    def set_law_z(self, expr: str):
        self.record_method('LawZ', expr)

    def set_param_range_u(self, u_min: float, u_max: float):
        self.record_method('ParameterRangeU', u_min, u_max)

    def set_param_range_v(self, v_min: float, v_max: float):
        self.record_method('ParameterRangeV', v_min, v_max)



from ..CST import FarfieldPlot
from .farfields import Farfield
import numpy as np

class FarfieldExporter:
    def __init__(self) -> None:
        self.coord_system = FarfieldPlot.COORD_SYSTEM_SPHERICAL
        self.plot_mode = FarfieldPlot.PLOT_MODE_EFIELD
        self.points = [] # array of tuples(polar angle, lateral angle, radius)
        self.current_farfield: Farfield = None
        self.origin = (0.0, 0.0, 0.0)

    def set_coord_system(self, coord_system: str):
        self.coord_system = coord_system

    def set_plot_mode(self, plot_mode: str):
        self.plot_mode = plot_mode

    def set_origin(self, x: float, y: float, z: float):
        self.origin = (x, y, z)

    def set_points(self, points):
        self.points = points

    def clear_points(self):
        self.points = []

    def add_point(self, polar_angle: float, lateral_angle: float, radius: float):
        self.points.append((polar_angle, lateral_angle, radius))

    def get_points(self):
        return self.points

    def prepare(self, farfield: Farfield):
        self.current_farfield = farfield
        self.current_farfield.select()
        ffplot = FarfieldPlot(farfield.project)
        ffplot.reset()
        ffplot.set_plot_mode(self.plot_mode)
        ffplot.set_plot_type(FarfieldPlot.PLOT_TYPE_3D)
        ffplot.set_db_unit(FarfieldPlot.UNIT_CODE_0) # basic units
        ffplot.set_origin(self.origin[0], self.origin[1], self.origin[2])
        ffplot.set_scale_linear()
        ffplot.plot()
        self.__generate_points()
        self.__add_plot_points(ffplot, farfield.get_frequency())
        ffplot.calculate_list()

    def export_abs(self):
        self.current_farfield.select()
        ffplot = FarfieldPlot(self.current_farfield.project)
        return np.asarray(ffplot.get_list(self.__get_comp_abs()))

    def export_complex_theta(self):
        self.current_farfield.select()
        ffplot = FarfieldPlot(self.current_farfield.project)
        re_comp = np.asarray(ffplot.get_list(self.__get_comp_theta_re()))
        im_comp = np.asarray(ffplot.get_list(self.__get_comp_theta_im()))
        return re_comp + 1j*im_comp

    def export_complex_phi(self):
        self.current_farfield.select()
        ffplot = FarfieldPlot(self.current_farfield.project)
        re_comp = np.asarray(ffplot.get_list(self.__get_comp_phi_re()))
        im_comp = np.asarray(ffplot.get_list(self.__get_comp_phi_im()))
        return re_comp + 1j*im_comp

    def __generate_points(self): # virtual method
        pass

    def __add_plot_points(self, ffplot: FarfieldPlot, frequency):
        for point in self.points:
            ffplot.add_list_eval_point_deg(
                point[0], point[1], point[2], self.coord_system,
                FarfieldPlot.TF_TYPE_FREQUENCY, frequency)

    def __get_comp_abs(self):
        return '{} {}'.format(self.coord_system, FarfieldPlot.COMPLEX_COMP_ABS)

    def __get_comp_theta_re(self):
        return '{} {} {} {}'.format(
            self.coord_system, FarfieldPlot.POLARIZATION_LINEAR,
            FarfieldPlot.COMPONENT_THETA, FarfieldPlot.COMPLEX_COMP_RE)

    def __get_comp_theta_im(self):
        return '{} {} {} {}'.format(
            self.coord_system, FarfieldPlot.POLARIZATION_LINEAR,
            FarfieldPlot.COMPONENT_THETA, FarfieldPlot.COMPLEX_COMP_IM)

    def __get_comp_phi_re(self):
        return '{} {} {} {}'.format(
            self.coord_system, FarfieldPlot.POLARIZATION_LINEAR,
            FarfieldPlot.COMPONENT_PHI, FarfieldPlot.COMPLEX_COMP_RE)

    def __get_comp_phi_im(self):
        return '{} {} {} {}'.format(
            self.coord_system, FarfieldPlot.POLARIZATION_LINEAR,
            FarfieldPlot.COMPONENT_PHI, FarfieldPlot.COMPLEX_COMP_IM)



from . import Project
from . import ComObjectWrapper

class MeshAdaption3D(ComObjectWrapper):
    TYPE_E_STATIC = 'EStatic'
    TYPE_M_STATIC = 'MStatic'
    TYPE_J_STATIC = 'JStatic'
    TYPE_LOW_FREQ = 'LowFrequency'
    TYPE_HIGH_FREQ_HEX = 'HighFrequencyHex'
    TYPE_HIGH_FREQ_TET = 'HighFrequencyTet'
    TYPE_TIME = 'Time'

    STRATEGY_EXPERT_SYSTEM = 'ExpertSystem'
    STRATEGY_ENERGY = 'Energy'

    REFINEMENT_TYPE_AUTOMATIC = 'Automatic'
    REFINEMENT_TYPE_BISECTION = 'Bisection'

    ERROR_ESTIMATOR_TYPE_AUTOMATIC = 'Automatic'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.MeshAdaption3D

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def set_error_limit(self, value: float):
        self.invoke_method('Errorlimit', value)

    def set_accuracy_factor(self, value: float):
        self.invoke_method('AccuracyFactor', value)

    def set_type(self, value: str):
        self.invoke_method('SetType', value)

    def set_adaptation_strategy(self, value: str):
        self.invoke_method('SetAdaptionStrategy', value)

    def set_refinement_type(self, value: str):
        self.invoke_method('RefinementType', value)

    def set_error_estimator_type(self, value: str):
        self.invoke_method('ErrorEstimatorType', value)

    def set_min_number_of_passes(self, value: int):
        self.invoke_method('MinPasses', value)

    def set_max_number_of_passes(self, value: int):
        self.invoke_method('MaxPasses', value)

    def set_cell_increase_factor(self, value: float):
        self.invoke_method('CellIncreaseFactor', value)

    def set_weight_e(self, value: float):
        self.invoke_method('WeightE', value)

    def set_weight_b(self, value: float):
        self.invoke_method('WeightB', value)

    def set_refine_x(self, flag: bool = True):
        self.invoke_method('RefineX', flag)

    def set_refine_y(self, flag: bool = True):
        self.invoke_method('RefineY', flag)

    def set_refine_z(self, flag: bool = True):
        self.invoke_method('RefineZ', flag)

    def set_refine_xyz(self, refine_x: bool, refine_y: bool, refine_z: bool):
        self.set_refine_x(refine_x)
        self.set_refine_y(refine_y)
        self.set_refine_z(refine_z)

    def set_max_delta_s(self, value: float):
        self.invoke_method('MaxDeltaS', value)

    def clear_stop_criteria(self):
        self.invoke_method('ClearStopCriteria')

    def add_s_param_stop_criterion(
            self, auto_freq: bool, f_min: float, f_max: float, max_delta: float,
            num_checks: int, active: bool = True):
        #TODO: split into multiple functions
        self.invoke_method(
            'AddSParameterStopCriterion', auto_freq, f_min, f_max, max_delta, num_checks, active)

    def add_0d_result_stop_criterion(
            self, result_name: str, max_delta: float, num_checks: int, active: bool = True):
        #TODO: split into multiple functions
        self.invoke_method('Add0DResultStopCriterion', result_name, max_delta, num_checks, active)

    def add_1d_result_stop_criterion(
            self, result_name: str, max_delta: float, num_checks: int, active: bool = True,
            is_complex: bool = True):
        #TODO: split into multiple functions
        self.invoke_method(
            'Add1DResultStopCriterion', result_name, max_delta, num_checks, active, is_complex)

    def add_stop_criterion(
            self, group_name: str, threshold: float, num_checks: int, active: bool = True):
        #TODO: split into multiple functions
        self.invoke_method('AddStopCriterion', group_name, threshold, num_checks, active)

    def remove_all_custom_stop_criteria(self):
        self.invoke_method('RemoveAllUserDefinedStopCriteria')

    def set_mesh_increment(self, value: float):
        self.invoke_method('MeshIncrement', value)

    def set_frequency_range(self, f_min: float, f_max: float):
        self.invoke_method('SetFrequencyRange', False, f_min, f_max)

    def set_frequency_range_auto(self):
        self.invoke_method('SetFrequencyRange', True, 0, 0)

    def set_number_of_skipped_pulses(self, value: float):
        self.invoke_method('SkipPulses', value)

    def set_number_of_delta_s_checks(self, value: int):
        self.invoke_method('NumberOfDeltaSChecks', value)

    def set_number_of_prop_const_checks(self, value: int):
        self.invoke_method('NumberOfPropConstChecks', value)

    def set_propagation_constant_accuracy(self, value: float):
        self.invoke_method('PropagationConstantAccuracy', value)

    def set_subsequent_checks_only_once(self, flag: bool = True):
        self.invoke_method('SubsequentChecksOnlyOnce', flag)

    def set_wavelength_based_refinement(self, flag: bool = True):
        self.invoke_method('WavelengthBasedRefinement', flag)

    def set_min_accepted_cell_growth(self, value: float):
        self.invoke_method('MinimumAcceptedCellGrowth', value)

    def set_ref_theta_factor(self, value: float):
        self.invoke_method('RefThetaFactor', value)

    def set_min_mesh_cell_growth(self, value: float):
        self.invoke_method('SetMinimumMeshCellGrowth', value)

    def set_linear_growth_limitation(self, value: float):
        self.invoke_method('SetLinearGrowthLimitation', value)

    def set_linear_growth_limitation_enabled(self, flag: bool = True):
        self.invoke_method('EnableLinearGrowthLimitation', flag)

    def set_singular_edge_refinement(self, value: int):
        self.invoke_method('SingularEdgeRefinement', value)

    def set_snap_to_geometry(self, flag: bool = True):
        self.invoke_method('SnapToGeometry', flag)

    def set_inner_s_param_adaptation_enabled(self, flag: bool = True):
        self.invoke_method('EnableInnerSParameterAdaptation', flag)

    def set_port_propagation_const_adaptation_enabled(self, flag: bool = True):
        self.invoke_method('EnablePortPropagationConstantAdaptation', flag)


from . import IVBAProvider, VBAObjWrapper, VBATypeName
import numpy as np

class PlaneWave(VBAObjWrapper):
    POLARIZATION_LINEAR = 'Linear'
    POLARIZATION_CIRCULAR = 'Circular'
    POLARIZATION_ELLIPTICAL = 'Elliptical'

    CIRCULAR_DIRECTION_LEFT = 'Left'
    CIRCULAR_DIRECTION_RIGHT = 'Right'

    DIRECTION_X = 'x'
    DIRECTION_Y = 'y'
    DIRECTION_Z = 'z'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'PlaneWave')

    def reset(self):
        self.record_method('Reset')

    def store(self):
        self.record_method('Store')

    def delete(self):
        self.record_method('Delete')

    def set_normal_vector(self, x: float, y: float, z: float):
        self.record_method('Normal', x, y, z)

    def set_e_vector(self, x: float, y: float, z: float):
        self.record_method('EVector', x, y, z)

    def set_polarization(self, polarization_type: str):
        self.record_method('Polarization', polarization_type)

    def set_reference_frequency(self, frequency: float):
        self.record_method('ReferenceFrequency', frequency)

    def set_phase_difference_deg(self, angle: float):
        self.record_method('PhaseDifference', angle)

    def set_phase_difference_rad(self, angle: float):
        self.set_phase_difference_deg(np.rad2deg(angle))

    def set_circular_direction(self, direction: str):
        self.record_method('CircularDirection', direction)

    def set_axial_ratio(self, ratio: float):
        self.record_method('AxialRatio', ratio)

    def set_use_custom_decoupling_plane(self, flag: bool = True):
        self.record_method('SetUserDecouplingPlane', flag)

    def set_custom_decoupling_plane(self, direction: str, position: float):
        self.record_method('DecouplingPlane', direction, position)

    # returns: (x, y, z)
    def get_normal_vector(self) -> tuple[float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetNormal', None, D, D, D)

    # returns: (x, y, z)
    def get_e_vector(self) -> tuple[float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetEVector', None, D, D, D)

    def get_polarization_type(self) -> str:
        S = VBATypeName.String
        return self.query_method_t('GetPolarizationType', None, S)[0]

    def get_circular_direction(self) -> str:
        S = VBATypeName.String
        return self.query_method_t('GetCircularDirection', None, S)[0]

    def get_reference_frequency(self) -> float:
        D = VBATypeName.Double
        return self.query_method_t('GetReferenceFrequency', None, D)[0]

    def get_phase_difference_deg(self) -> float:
        D = VBATypeName.Double
        return self.query_method_t('GetPhaseDifference', None, D)[0]

    def get_phase_difference_rad(self) -> float:
        return np.deg2rad(self.get_phase_difference_deg())

    def get_axial_ratio(self) -> float:
        D = VBATypeName.Double
        return self.query_method_t('GetAxialRatio', None, D)[0]



from . import IVBAProvider, VBAObjWrapper

class ParameterSweep(VBAObjWrapper):
    SIMULATION_TYPE_TRANSIENT = 'Transient'
    SIMULATION_TYPE_PORT_MODES_ONLY = 'Calculate port modes only'
    SIMULATION_TYPE_EIGENMODE = 'Eigenmode'
    SIMULATION_TYPE_FREQUENCY = 'Frequency'
    SIMULATION_TYPE_TLM = 'TLM'
    SIMULATION_TYPE_ASYMPTOTIC = 'Asymtotic'

    SIMULATION_TYPE_E_STATIC = 'E-Static'
    SIMULATION_TYPE_ELECTROQUASISTATIC = 'Electroquasistatic'
    SIMULATION_TYPE_TRANSIENT_ELECTROQUASISTATIC = 'Transient Electroquasistatic'
    SIMULATION_TYPE_M_STATIC = 'M-Static'
    SIMULATION_TYPE_MAGNETOQUASISTATIC = 'Transient Magnetoquasistatic'
    SIMULATION_TYPE_J_STATIC = 'J-Static'
    SIMULATION_TYPE_LOW_FREQUENCY = 'Low Frequency'

    SIMULATION_TYPE_THERMAL = 'Thermal'
    SIMULATION_TYPE_TRANSIENT_THERMAL = 'Transient Thermal'
    SIMULATION_TYPE_STRUCTURAL_MECHANICS = 'Structural Mechanics'

    SIMULATION_TYPE_PIC = 'PIC'
    SIMULATION_TYPE_PARTICLE_TRACKING = 'Particle Tracking'

    #TODO: check which methods should be recorded and which invoked

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'ParameterSweep')

    def set_simulation_type(self, sim_type: str):
        self.invoke_method('SetSimulationType', sim_type)

    def add_sequence(self, name: str):
        self.invoke_method('AddSequence', name)

    def delete_sequence(self, name: str):
        self.invoke_method('DeleteSequence', name)

    def set_sequence_enabled(self, name: str, flag: bool = True):
        self.invoke_method('EnableSequence', name, flag)

    def delete_all_sequences(self):
        self.invoke_method('DeleteAllSequences')

    def rename_sequence(self, old_name: str, new_name: str):
        self.invoke_method('RenameSequence', old_name, new_name)

    def add_parameter_samples_lin(
            self, seq_name: str, param_name: str, lower_bound: float, upper_bound: float,
            num_steps: int):
        self.invoke_method(
            'AddParameter_Samples', seq_name, param_name, lower_bound, upper_bound,
            num_steps, False)

    def add_parameter_samples_log(
            self, seq_name: str, param_name: str, lower_bound: float, upper_bound: float,
            num_steps: int):
        self.invoke_method(
            'AddParameter_Samples', seq_name, param_name, lower_bound, upper_bound,
            num_steps, True)

    def add_parameter_step_width(
            self, seq_name: str, param_name: str, lower_bound: float, upper_bound: float,
            width: float):
        self.invoke_method(
            'AddParameter_Stepwidth', seq_name, param_name, lower_bound, upper_bound, width)

    def add_parameter_arbitrary_points(self, seq_name: str, param_name: str, points: list[float]):
        points_str = ';'.join(str(s) for s in points)
        self.invoke_method('AddParameter_ArbitraryPoints', seq_name, param_name, points_str)

    def delete_parameter(self, seq_name: str, param_name: str):
        self.invoke_method('DeleteParameter', seq_name, param_name)

    def start(self):
        self.invoke_method('Start')

    def set_use_distributed_computing(self, flag: bool = True):
        self.invoke_method('UseDistributedComputing', flag)


from . import IVBAProvider, VBAObjWrapper, VBATypeName
from typing import Optional

class Anchorpoint(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Anchorpoint')

    def store(self, name: str):
        self.record_method('Store', name)

    def define(
            self, name: str,
            pos_x: float, pos_y: float, pos_z: float,
            normal_x: float, normal_y: float, normal_z: float,
            u_vector_x: float, u_vector_y: float, u_vector_z: float):
        self.record_method(
            'Define', name,
            pos_x, pos_y, pos_z,
            normal_x, normal_y, normal_z,
            u_vector_x, u_vector_y, u_vector_z)

    def restore(self, name: str):
        self.record_method('Restore', name)

    def delete(self, name: str):
        self.record_method('Delete', name)

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def create_folder(self, name: str):
        self.record_method('NewFolder', name)

    def delete_folder(self, name: str):
        self.record_method('DeleteFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.record_method('RenameFolder', old_name, new_name)

    def does_exist(self, name: str) -> bool:
        return self.query_method_bool('DoesExist', name)

    # returns: (x, y, z)
    def get_origin(self, name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetOrigin', B, name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (x, y, z)
    def get_normal(self, name: str):
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetNormal', B, name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (x, y, z)
    def get_u_vector(self, name: str):
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetUVector', B, name, D, D, D)
        return None if not res[0] else res[1:]


from . import Project
from . import Shape

class Brick(Shape):
    def __init__(self, project: Project) -> None:
        super().__init__(project, project.com_object.Brick)

    def set_x_range(self, x_min: float, x_max: float):
        self.invoke_method('Xrange', x_min, x_max)

    def set_y_range(self, y_min: float, y_max: float):
        self.invoke_method('Yrange', y_min, y_max)

    def set_z_range(self, z_min: float, z_max: float):
        self.invoke_method('Zrange', z_min, z_max)


from win32com.client import VARIANT
import pythoncom

def create_ref_double() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_R8, -1)

def create_ref_long() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_I8, -1)

def create_ref_int() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_I4, -1)

def create_ref_str() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_BSTR, -1)

def create_ref_bool() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_BOOL, -1)

def create_ref_variant() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_VARIANT, -1)

def create_ref_variant_array() -> VARIANT:
    return VARIANT(pythoncom.VT_BYREF | pythoncom.VT_ARRAY | pythoncom.VT_VARIANT, -1)

def create_str_array(arr: list[any]) -> VARIANT:
    arr = list(map(lambda x: str(x), arr))
    return VARIANT(pythoncom.VT_ARRAY | pythoncom.VT_BSTR, arr)


from . import IVBAProvider, VBAObjWrapper

class Optimizer(VBAObjWrapper):
    GOAL_TYPE_1D_PRIMARY_RESULT = '1D Primary Result'
    GOAL_TYPE_1DC_PRIMARY_RESULT = '1DC Primary Result'
    GOAL_TYPE_0D_RESULT = '0D Result'
    GOAL_TYPE_1D_RESULT = '1D Result'
    GOAL_TYPE_1DC_RESULT = '1DC Result'

    GOAL_SUMMARY_TYPE_SUM_ALL_GOALS = 'Sum_All_Goals'
    GOAL_SUMMARY_TYPE_MAX_ALL_GOALS = 'Max_All_Goals'

    GOAL_OPERATOR_LESS_THAN = '<'
    GOAL_OPERATOR_GREATER_THAN = '>'
    GOAL_OPERATOR_EQUAL = '='
    GOAL_OPERATOR_MIN = 'min'
    GOAL_OPERATOR_MAX = 'max'
    GOAL_OPERATOR_MOVE_MIN = 'move min'
    GOAL_OPERATOR_MOVE_MAX = 'move max'

    GOAL_NORM_MAX_DIFF = 'MaxDiff'
    GOAL_NORM_MAX_DIFF_SQUARED = 'MaxDiffSq'
    GOAL_NORM_SUM_DIFF = 'SumDiff'
    GOAL_NORM_SUM_DIFF_SQUARED = 'SumDiffSq'
    GOAL_NORM_DIFF = 'Diff'
    GOAL_NORM_DIFF_SQUARED = 'DiffSq'

    GOAL_SCALAR_TYPE_MAG_LIN = 'maglin'
    GOAL_SCALAR_TYPE_MAG_DB_10 = 'magdb10'
    GOAL_SCALAR_TYPE_MAG_DB_20 = 'magdb20'
    GOAL_SCALAR_TYPE_REAL = 'real'
    GOAL_SCALAR_TYPE_IMAG = 'imag'
    GOAL_SCALAR_TYPE_PHASE = 'phase'

    GOAL_RANGE_TYPE_TOTAL = 'total'
    GOAL_RANGE_TYPE_RANGE = 'range'
    GOAL_RANGE_TYPE_SINGLE = 'single'

    OPTIMIZER_TYPE_TRUST_REGION = 'Trust_Region'
    OPTIMIZER_TYPE_NELDER_MEAD_SIMPLEX = 'Nelder_Mead_Simplex'
    OPTIMIZER_TYPE_CMAES = 'CMAES'
    OPTIMIZER_TYPE_GENETIC_ALGORITHM = 'Genetic_Algorithm'
    OPTIMIZER_TYPE_PARTICLE_SWARM = 'Particle_Swarm'
    OPTIMIZER_TYPE_INTERPOLATED_NR_VAR_METRIC = 'Interpolated_NR_VariableMetric'
    OPTIMIZER_TYPE_CLASSIC_POWELL = 'Classic Powell'

    INTERPOLATION_SECOND_ORDER = 'Second_Order'

    DISTRIBUTION_TYPE_UNIFORM_RANDOM_NUMBERS = 'Uniform_Random_Numbers'
    DISTRIBUTION_TYPE_LATIN_HYPER_CUBE = 'Latin_Hyper_Cube'
    DISTRIBUTION_TYPE_NOISY_LATIN_HYPER_CUBE = 'Noisy_Latin_Hyper_Cube'
    DISTRIBUTION_TYPE_CUBE = 'Cube_Distribution'

    DATA_STORAGE_STRATEGY_ALL = 'All'
    DATA_STORAGE_STRATEGY_AUTOMATIC = 'Automatic'
    DATA_STORAGE_STRATEGY_NONE = 'None'

    #TODO: check which methods should be recorded and which invoked

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Optimizer')

    def start(self):
        self.invoke_method('Start')

    def set_start_active_solver(self, flag: bool = True):
        self.invoke_method('StartActiveSolver', flag)

    def init_parameter_list(self):
        self.invoke_method('InitParameterList')

    def reset_parameter_list(self):
        self.invoke_method('ResetParameterList')

    def select_parameter(self, param_name: str, use_for_optimization: bool = True):
        self.invoke_method('SelectParameter', param_name, use_for_optimization)

    def set_parameter_init_value(self, value: float):
        self.invoke_method('SetParameterInit', value)

    def set_parameter_min_value(self, value: float):
        self.invoke_method('SetParameterMin', value)

    def set_parameter_max_value(self, value: float):
        self.invoke_method('SetParameterMax', value)

    def set_parameter_number_of_anchors(self, number: int):
        self.invoke_method('SetParameterAnchors', number)

    def set_min_max_auto(self, percentage: float):
        self.invoke_method('SetMinMaxAuto', percentage)

    def set_and_update_min_max_auto(self, percentage: float):
        self.invoke_method('SetAndUpdateMinMaxAuto', percentage)

    def set_always_start_from_current(self, flag: bool = True):
        self.invoke_method('SetAlwaysStartFromCurrent', flag)

    def set_use_data_of_previous_calculations(self, flag: bool = True):
        self.invoke_method('SetUseDataOfPreviousCalculations', flag)

    def get_number_of_varying_parameters(self) -> int:
        return self.query_method_int('GetNumberOfVaryingParameters')

    def get_name_of_varying_parameter(self, index: int) -> str:
        return self.query_method_str('GetNameOfVaryingParameter', index)

    def get_value_of_varying_parameter(self, index: int) -> float:
        return self.query_method_float('GetValueOfVaryingParameter', index)

    def get_param_min_of_varying_parameter(self, index: int) -> float:
        return self.query_method_float('GetParameterMinOfVaryingParameter', index)

    def get_param_max_of_varying_parameter(self, index: int) -> float:
        return self.query_method_float('GetParameterMaxOfVaryingParameter', index)

    def get_param_init_of_varying_parameter(self, index: int) -> float:
        return self.query_method_float('GetParameterInitOfVaryingParameter', index)

    def add_goal(self, goal_type: str) -> int:
        return self.query_method_int('AddGoal', goal_type)

    def select_goal(self, id: int, use_for_optimization: bool = True):
        self.invoke_method('SelectGoal', id, use_for_optimization)

    def delete_goal(self, id: int):
        self.invoke_method('DeleteGoal', id)

    def delete_all_goals(self):
        self.invoke_method('DeleteAllGoals')

    def set_gial_summary_type(self, goal_summary_type: str):
        self.invoke_method('SetGoalSummaryType', goal_summary_type)

    def set_use_goal_for_optimization(self, flag: bool = True):
        self.invoke_method('SetGoalUseFlag', flag)

    def set_goal_operator(self, operator_type: str):
        self.invoke_method('SetGoalOperator', operator_type)

    def set_goal_target(self, value: float):
        self.invoke_method('SetGoalTarget', value)

    def set_goal_norm(self, norm: str):
        self.invoke_method('SetGoalNormNew', norm)

    def set_goal_weight(self, value: float):
        self.invoke_method('SetGoalWeight', value)

    def set_goal_scalar_type(self, scalar_type: str):
        self.invoke_method('SetGoalScalarType', scalar_type)

    def set_goal_1d_result_name(self, name: str):
        self.invoke_method('SetGoal1DResultName', name)

    def set_goal_1dc_result_name(self, name: str):
        self.invoke_method('SetGoal1DCResultName', name)

    def set_goal_template_based_0d_result_name(self, name: str):
        self.invoke_method('SetGoalTemplateBased0DResultName', name)

    def set_goal_template_based_1d_result_name(self, name: str):
        self.invoke_method('SetGoalTemplateBased1DResultName', name)

    def set_goal_template_based_1dc_result_name(self, name: str):
        self.invoke_method('SetGoalTemplateBased1DCResultName', name)

    def set_goal_range(self, min_value: float, max_value: float):
        self.invoke_method('SetGoalRange', min_value, max_value)

    def set_goal_range_type(self, range_type: str):
        self.invoke_method('SetGoalRangeType', range_type)

    def set_use_slope(self, flag: bool = True):
        self.invoke_method('UseSlope', flag)

    def set_goal_target_max(self, value: float):
        self.invoke_method('SetGoalTargetMax', value)

    def set_optimizer_type(self, optimizer_type: str):
        self.invoke_method('SetOptimizerType', optimizer_type)

    def set_use_2nd_order_interpolation(self, optimizer_type: str):
        self.invoke_method('SetUseInterpolation', self.INTERPOLATION_SECOND_ORDER, optimizer_type)

    def set_generation_size(self, optimizer_type: str, size: float):
        self.invoke_method('SetGenerationSize', size, optimizer_type)

    def set_max_number_of_iterations(self, optimizer_type: str, number: int):
        self.invoke_method('SetMaxIt', number, optimizer_type)

    def set_initial_distribution(self, optimizer_type: str, distribution: str):
        self.invoke_method('SetInitialDistribution', distribution, optimizer_type)

    def set_goal_function_level(self, optimizer_type: str, level: float):
        self.invoke_method('SetGoalFunctionLevel', level, optimizer_type)

    def set_goal_function_level(self, optimizer_type: str, rate: float):
        #TODO: in the docs there is 'SetMutaionRate', check whether it is really a mistake
        self.invoke_method('SetMutationRate', rate, optimizer_type)

    def set_min_simplex_size(self, value: float):
        self.invoke_method('SetMinSimplexSize', value)

    def set_use_max_number_of_evaluations(self, optimizer_type: str, flag: bool = True):
        self.invoke_method('SetUseMaxEval', flag, optimizer_type)

    def set_max_number_of_evaluations(self, optimizer_type: str, number: int):
        self.invoke_method('SetMaxEval', number, optimizer_type)

    def set_use_pre_def_point_in_init_distribution(self, optimizer_type: str, flag: bool = True):
        self.invoke_method('SetUsePreDefPointInInitDistribution', flag, optimizer_type)

    def set_number_of_refinements(self, number: int):
        self.invoke_method('SetNumRefinements', number)

    def set_domain_accuracy(self, optimizer_type: str, accuracy: float):
        self.invoke_method('SetDomainAccuracy', accuracy, optimizer_type)

    def set_sigma(self, optimizer_type: str, sigma: float):
        self.invoke_method('SetSigma', sigma, optimizer_type)

    def set_accuracy(self, optimizer_type: str, accuracy: float):
        self.invoke_method('SetAccuracy', accuracy, optimizer_type)

    def set_data_storage_strategy(self, strategy: str):
        self.invoke_method('SetDataStorageStrategy', strategy)

    def set_move_mesh(self, flag: bool = True):
        self.invoke_method('SetOptionMoveMesh', flag)


__ALL__ = [
]

from . import info, units
from .frequency_range_list import FrequencyRangeList
from .vba_type_name import VBATypeName
from .ivba_runtime import IVBARuntime
from .iquiet_mode_controller import IQuietModeController
from .ihistory_list_provider import IHistoryListProvider
from .history_list_generator import IHistoryListGeneratorProvider
from .history_list_generator import HistoryListGenerator
from .ivba_provider import IVBAProvider
from .modeler import Modeler
from .schematic import Schematic
from .interface_pcbs import InterfacePCBS
from .vba_obj_wrapper import VBAObjWrapper
from .parameters import Parameters
from .iproject import IProject
from .shape import Shape
from .ads_cosimulation import ADSCosimulation
from .align import Align
from .analytical_curve import AnalyticalCurve
from .anchorpoint import Anchorpoint
from .arc import Arc
from .asymptotic_solver import AsymptoticSolver
from .background import Background
from .boundary import Boundary
from .brick import Brick
from .component import Component
from .cone import Cone
from .curve import Curve
from .cylinder import Cylinder
from .port import Port
from .discrete_face_port import DiscreteFacePort
from .discrete_port import DiscretePort
from .eigenmode_solver import EigenmodeSolver
from .elliptical_cylinder import EllipticalCylinder
from .extrude import Extrude
from .farfield_plot import FarfieldPlot
from .farfield_source import FarfieldSource
from .fd_solver import FDSolver
from .field_source import FieldSource
from .floquet_port import FloquetPort
from .group import Group
from .hf_solver import HFSolver
from .ie_solver import IESolver
from .layer_stacking import LayerStacking
from .loft import Loft
from .lumped_element import LumpedElement
from .lumped_face_element import LumpedFaceElement
from .material import Material
from .mesh_adaptation_3d import MeshAdaption3D
from .mesh_settings import MeshSettings
from .mesh_shapes import MeshShapes
from .mesh import Mesh
from .monitor import Monitor
from .optimizer import Optimizer
from .parameter_sweep import ParameterSweep
from .pick import Pick
from .plane_wave import PlaneWave
from .post_process_1d import PostProcess1D
from .project_units import ProjectUnits
from .rotate import Rotate
from .solid import Solid
from .solver_parameter import SolverParameter
from .sphere import Sphere
from .stl import STL
from .time_signal import TimeSignal
from .torus import Torus
from .trace_from_curve import TraceFromCurve
from .transform import Transform
from .wcs import WCS
from .wire import Wire
from .result_item import ResultItem
from .result_module import ResultModule
from .results import Results
from .project import Project
from .interface import Interface
# from .result_tree import ResultTree
# from .result_tree_item import ResultTreeItem
# from .farfield_plot import FarfieldPlot
# from .result_1d import Result1D
# from .result_1d_complex import Result1DComplex


from . import Project
from . import ComObjectWrapper

class SolverParameter(ComObjectWrapper):
    SOLVER_TYPE_TRANSIENT = 'Transient solver'
    SOLVER_TYPE_FREQ_DOMAIN = 'Frequency domain solver'
    SOLVER_TYPE_ASYMPTOTIC = 'Asymptotic solver'
    SOLVER_TYPE_EIGENMODE = 'Eigenmode solver'
    SOLVER_TYPE_ELECTROSTATICS = 'Electrostatics solver'
    SOLVER_TYPE_MAGNETOSTATIC = 'Magnetostatic solver'
    SOLVER_TYPE_LF_FREQ_DOMAIN = 'LF Frequency domain solver'
    SOLVER_TYPE_LF_FREQ_DOMAIN_EQS = 'LF Frequency domain solver (EQS)'
    SOLVER_TYPE_STATIONARY_CURRENT = 'Stationary current solver'
    SOLVER_TYPE_PARTICLE_TRACKING = 'Particle tracking solver'
    SOLVER_TYPE_PIC = 'PIC solver'
    SOLVER_TYPE_THERMAL = 'Thermal solver'
    SOLVER_TYPE_INTEGRAL_EQUATION = 'Integral equation solver'
    SOLVER_TYPE_MULTILAYER = 'Multilayer solver'
    SOLVER_TYPE_LF_TIME_DOMAIN_MQS = 'LF Time domain solver (MQS)'
    SOLVER_TYPE_LF_TIME_DOMAIN_EQS = 'LF Time domain solver (EQS)'
    SOLVER_TYPE_THERMAL_TRANSIENT = 'Thermal transient solver'
    SOLVER_TYPE_STRUCTURAL_MECHANICS = 'Structural mechanics solver'
    SOLVER_TYPE_WAKEFIELD = 'Wakefield solver'

    MESH_HEXAHEDRAL = 'Hexahedral'
    MESH_TETRAHEDRAL = 'Tetrahedral'
    MESH_SURFACE = 'Surface'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.SolverParameter

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def set_solver_type(self, solver_type: str, mesh: str):
        self.invoke_method('SolverType', solver_type, mesh)

    def set_ignore_lossy_metals(self, flag: bool = True):
        self.invoke_method('IgnoreLossyMetals', flag)

    def set_ignore_lossy_dielectrics(self, flag: bool = True):
        self.invoke_method('IgnoreLossyDielectrics', flag)

    def set_ignore_lossy_metals_for_wires(self, flag: bool = True):
        self.invoke_method('IgnoreLossyMetalsForWires', flag)

    def set_ignore_nonlinear_materials(self, flag: bool = True):
        self.invoke_method('IgnoreNonlinearMaterials', flag)

    def set_use_thin_wire_model(self, flag: bool = True):
        self.invoke_method('UseThinWireModel', flag)

    def set_use_zero_wire_radius(self, flag: bool = True):
        self.invoke_method('UseZeroWireRadius', flag)


from . import Project
from . import ComObjectWrapper

class TimeSignal(ComObjectWrapper):
    TYPE_GAUSSIAN = 'Gaussian'
    TYPE_RECTANGULAR = 'Rectangular'
    TYPE_SINE_STEP = 'Sine step'
    TYPE_SINE = 'Sine'
    TYPE_SMOOTH_STEP = 'Smooth step'
    TYPE_CONSTANT = 'Constant'
    TYPE_DOUBLE_EXPONENTIAL = 'Double exponential'
    TYPE_IMPULSE = 'Impulse'
    TYPE_USER = 'User'
    TYPE_IMPORT = 'Import'

    PROBLEM_TYPE_HIGH_FREQUENCY = 'High Frequency'
    PROBLEM_TYPE_LOW_FREQUENCY = 'Low Frequency'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.TimeSignal

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_id(self, id: int):
        self.invoke_method('Id', id)

    def rename(self, old_name: str, new_name: str, problem_type: str):
        self.invoke_method('Rename', old_name, new_name, problem_type)

    def delete(self, name: str, problem_type: str):
        self.invoke_method('Delete', name, problem_type)

    def create(self):
        self.invoke_method('Create')

    def set_file_name(self, file_name: str):
        self.invoke_method('FileName', file_name)

    def set_use_copy_only(self, flag: bool = True):
        self.invoke_method('UseCopyOnly', flag)

    def set_f_min(self, value: float):
        self.invoke_method('Fmin', value)

    def set_f_max(self, value: float):
        self.invoke_method('Fmax', value)

    def set_total_time(self, time: float):
        self.invoke_method('Ttotal', time)

    def set_rise_time(self, time: float):
        self.invoke_method('Trise', time)

    def set_hold_time(self, time: float):
        self.invoke_method('Thold', time)

    def set_fall_time(self, time: float):
        self.invoke_method('Tfall', time)

    def set_vertical_offset(self, offset: float):
        self.invoke_method('Voffset', offset)

    def set_amplitude_rise_percent(self, percent: float):
        self.invoke_method('AmplitudeRisePercent', percent)

    def set_rise_factor(self, factor: float):
        self.invoke_method('RiseFactor', factor)

    def set_chirp_rate(self, rate: float):
        self.invoke_method('ChirpRate', rate)

    def set_frequency(self, frequency: float):
        self.invoke_method('Frequency', frequency)

    def set_phase(self, phase: float):
        self.invoke_method('Phase', phase)

    def set_amplitude(self, amplitude: float):
        self.invoke_method('Amplitude', amplitude)

    def set_start_amplitude(self, amplitude: float):
        self.invoke_method('StartAmplitude', amplitude)

    def set_end_amplitude(self, amplitude: float):
        self.invoke_method('EndAmplitude', amplitude)

    def set_signal_type(self, signal_type: str):
        self.invoke_method('SignalType', signal_type)

    def set_min_number_of_user_signal_samples(self, number: int):
        self.invoke_method('MinUserSignalSamples', number)

    def set_periodic(self, flag: bool = True):
        self.invoke_method('Periodic', flag)

    def set_problem_type(self, problem_type: str):
        self.invoke_method('ProblemType', problem_type)

    def set_excitation_signal_as_reference(self, signal_name: str, problem_type: str):
        self.invoke_method('ExcitationSignalAsReference', signal_name, problem_type)

    def resample_excitation_signal(
            self, signal_name: str, t_min: float, t_max: float, t_step: float, problem_type: str):
        self.invoke_method(
            'ExcitationSignalResample', signal_name, t_min, t_max, t_step, problem_type)

    def get_next_id(self) -> int:
        return self.invoke_method('GetNextId')


from typing import Optional

class Schematic:
    def __init__(self, native_obj) -> None:
        self.native_obj = native_obj

    def execute_vba_code(self, vba_code: str) -> None:
        self.native_obj.execute_vba_code(vba_code)


from .. import installation_util
from . import Project, IQuietModeController
from typing import Optional, Union
from enum import Enum
import os.path

class Interface(IQuietModeController):
    class StartMode(Enum):
        import cst.interface
        New = cst.interface.DesignEnvironment.StartMode.New
        ExistingOrNew = cst.interface.DesignEnvironment.StartMode.ExistingOrNew
        Existing = cst.interface.DesignEnvironment.StartMode.Existing

    def __init__(
            self, version_or_install_dir: Optional[Union[int,str]] = None,
            start_mode: StartMode = StartMode.ExistingOrNew) -> None:
        installation_util.load_win_cst_python_lib(version_or_install_dir)
        self.stored_quiet_mode = False
        self.__init_cst_interface(start_mode)
        self.set_quiet_mode(True)

    def __del__(self) -> None:
        # FIXME: here should be set_quiet_mode(False) but it doesn't work here (for an unknown
        # reason the connection to the Design Environment is lost)
        pass

    def __init_cst_interface(self, start_mode: StartMode = StartMode.New) -> None:
        import cst.interface
        self.design_env = cst.interface.DesignEnvironment(mode=start_mode.value)

    def store_quiet_mode(self) -> None:
        self.stored_quiet_mode = self.is_in_quiet_mode()

    def restore_quiet_mode(self) -> None:
        self.set_quiet_mode(self.stored_quiet_mode)

    def store_and_disable_quiet_mode(self) -> None:
        self.store_quiet_mode()
        self.set_quiet_mode(False)

    def store_and_enable_quiet_mode(self) -> None:
        self.store_quiet_mode()
        self.set_quiet_mode(True)

    def set_quiet_mode(self, enabled: bool = True) -> None:
        self.design_env.set_quiet_mode(enabled)

    def is_in_quiet_mode(self) -> bool:
        return self.design_env.in_quiet_mode()

    def get_open_projects(self) -> list[str]:
        return self.design_env.list_open_projects()

    def open_project(self, path: str) -> Project:
        return Project(self.design_env.open_project(os.path.abspath(path)), self)

    def get_active_project(self) -> Project:
        return Project(self.design_env.active_project(), self)

    def new_project(self, kind: Project.Kind) -> Project:
        return Project(self.design_env.new_project(kind.value), self)

    def new_cable_studio_project(self) -> Project:
        return Project(self.design_env.new_cs(), self)

    def new_design_studio_project(self) -> Project:
        return Project(self.design_env.new_ds(), self)

    def new_em_studio_project(self) -> Project:
        return Project(self.design_env.new_ems(), self)

    def new_filter_designer_3d_project(self) -> Project:
        return Project(self.design_env.new_fd3d(), self)

    def new_mphysics_studio_project(self) -> Project:
        return Project(self.design_env.new_mps(), self)

    def new_microwave_studio_project(self) -> Project:
        return Project(self.design_env.new_mws(), self)

    def new_pcb_studio_project(self) -> Project:
        return Project(self.design_env.new_pcbs(), self)

    def new_particle_studio_project(self) -> Project:
        return Project(self.design_env.new_ps(), self)

    def close(self) -> None:
        self.design_env.close()

    def get_version() -> str:
        import cst.interface
        return cst.interface.DesignEnvironment.version()


from . import Project
from . import ComObjectWrapper
from . import w32com
import numpy as np

class Port(ComObjectWrapper):
    ORIENTATION_XMIN = 'xmin'
    ORIENTATION_XMAX = 'xmax'
    ORIENTATION_YMIN = 'ymin'
    ORIENTATION_YMAX = 'ymax'
    ORIENTATION_ZMIN = 'zmin'
    ORIENTATION_ZMAX = 'zmax'

    COORDINATES_FREE = 'Free'
    COORDINATES_FULL = 'Full'
    COORDINATES_PICKS = 'Picks'

    POTENTIAL_POSITIVE = 'Positive'
    POTENTIAL_NEGATIVE = 'Negative'

    BOUNDARY_POS_UMIN = 'Umin'
    BOUNDARY_POS_UMAX = 'Umax'
    BOUNDARY_POS_VMIN = 'Vmin'
    BOUNDARY_POS_VMAX = 'Vmax'

    SHIELD_NONE = 'none'
    SHIELD_PEC = 'PEC'
    SHIELD_PMC = 'PMC'

    MODE_TE = 'TE'
    MODE_TM = 'TM'
    MODE_TEM = 'TEM'
    MODE_QTEM = 'QTEM'
    MODE_UNDEF = 'UNDEF'
    MODE_DAMPED = 'DAMPED'
    MODE_PLANE_WAVE = 'PLANE WAVE'
    MODE_FLOQUET = 'FLOQUET'

    FACE_PORT_TYPE_UNKNOWN = 0
    FACE_PORT_TYPE_RECTANGULAR = 1
    FACE_PORT_TYPE_CYLINDRICAL = 2
    FACE_PORT_TYPE_COAXIAL = 3

    def __init__(self, project) -> None:
        self.project: Project = project
        self.com_object = project.com_object.Port

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def delete(self, port_number: int):
        self.invoke_method('Delete', port_number)

    def change_number(self, old_port_number: int, new_port_number: int):
        self.invoke_method('Rename', old_port_number, new_port_number)

    def change_label(self, port_number: int, label: str):
        self.invoke_method('RenameLabel', port_number, label)

    def start_port_number_iteration(self) -> int:
        return self.invoke_method('StartPortNumberIteration')

    def start_sparam_port_number_iteration(self) -> int:
        return self.invoke_method('StartSParaPortNumberIteration')

    def get_next_port_number(self) -> int:
        return self.invoke_method('GetNextPortNumber')

    def load_content_for_modify(self) -> int:
        return self.invoke_method('LoadContentForModify')

    def create(self):
        self.invoke_method('Create')

    def modify(self):
        self.invoke_method('Modify')

    def set_number(self, number: int):
        self.invoke_method('PortNumber', number)

    def set_label(self, label: str):
        self.invoke_method('Label', label)

    def set_folder_name(self, folder_name: str):
        # undocumented, found in the history list
        self.invoke_method('Folder', folder_name)

    def set_waveguide_monitor(self, flag: bool = True):
        # undocumented, found in the history list
        self.invoke_method('WaveguideMonitor', flag)

    def set_number_of_modes(self, count: int):
        self.invoke_method('NumberOfModes', count)

    def set_orientation(self, orientation: str):
        self.invoke_method('Orientation', orientation)

    def set_coordinates(self, coordinates: str):
        self.invoke_method('Coordinates', coordinates)

    def set_x_range(self, x_min: float, x_max: float):
        self.invoke_method('Xrange', x_min, x_max)

    def set_y_range(self, y_min: float, y_max: float):
        self.invoke_method('Yrange', y_min, y_max)

    def set_z_range(self, z_min: float, z_max: float):
        self.invoke_method('Zrange', z_min, z_max)

    def set_x_position(self, x: float):
        self.set_x_range(x, x)

    def set_y_position(self, y: float):
        self.set_y_range(y, y)

    def set_z_position(self, z: float):
        self.set_z_range(z, z)

    def add_x_range(self, x_min: float, x_max: float):
        self.invoke_method('XrangeAdd', x_min, x_max)

    def add_y_range(self, y_min: float, y_max: float):
        self.invoke_method('YrangeAdd', y_min, y_max)

    def add_z_range(self, z_min: float, z_max: float):
        self.invoke_method('ZrangeAdd', z_min, z_max)

    def set_on_boundaries(self, flag: bool = True):
        self.invoke_method('PortOnBound', flag)

    def set_clip_picked_port_to_boundaries(self, flag: bool = True):
        self.invoke_method('ClipPickedPortToBound', flag)

    def set_text_size(self, size: int):
        self.invoke_method('TextSize', size)

    def change_text_size(self, port_number: int, size: int):
        self.invoke_method('ChangeTextSize', port_number, size)

    def set_text_max_limit(self, flag: bool = True):
        self.invoke_method('TextMaxLimit', flag)

    def set_reference_plane_distance(self, distance: float):
        self.invoke_method('ReferencePlaneDistance', distance)

    def add_potential_numerically(self, mode_set: int, potential: str, u_pos: float, v_pos: float):
        self.invoke_method('AddPotentialNumerically', mode_set, potential, u_pos, v_pos)

    def add_potential_picked(self, mode_set: int, potential: str, solid_name: str, face_id: int):
        self.invoke_method('AddPotentialPicked', mode_set, potential, solid_name, face_id)

    def add_potential_edge_picked(
            self, mode_set: int, potential: str, solid_name: str, edge_id: int):
        self.invoke_method('AddPotentialEdgePicked', mode_set, potential, solid_name, edge_id)

    def set_adjust_polarization(self, flag: bool = True):
        self.invoke_method('AdjustPolarization', flag)

    def set_polarization_angle_deg(self, angle_deg: float):
        self.invoke_method('PolarizationAngle', angle_deg)

    def set_polarization_angle_rad(self, angle_rad: float):
        self.set_polarization_angle_deg(np.rad2deg(angle_rad))

    def set_port_impedance_and_calibration_evaluation(self, flag: bool = True):
        self.invoke_method('PortImpedanceAndCalibration', flag)

    def add_mode_line_by_point(
            self, line_number: int,
            x_start: float, y_start: float, z_start: float,
            x_end: float, y_end: float, z_end: float):
        self.invoke_method(
            'AddModeLineByPoint', line_number, x_start, y_start, z_start, x_end, y_end, z_end)

    def add_mode_line_by_face(
            self, line_number: int, x_start: float, y_start: float, z_start: float,
            solid_name: str, face_id: int, reverse: bool = False):
        self.invoke_method(
            'AddModeLineByFace', line_number,
            x_start, y_start, z_start, solid_name, face_id, reverse)

    def add_mode_line_by_boundary(
            self, line_number: int, x_start: float, y_start: float, z_start: float,
            position: str, reverse: bool = False):
        self.invoke_method(
            'AddModeLineByBoundary', line_number, x_start, y_start, z_start, position, reverse)

    def add_mode_line(
            self, mode_number: int, impedance_line_number: int, calibration_line_number: int,
            polarization_line_number: int):
        self.invoke_method(
            'AddModeLine', mode_number, impedance_line_number, calibration_line_number,
            polarization_line_number)

    def set_estimation(self, port_number: int, value: float):
        self.invoke_method('SetEstimation', port_number, value)

    def set_single_ended(self, flag: bool = True):
        self.invoke_method('SingleEnded', flag)

    def set_shield(self, key: str):
        self.invoke_method('Shield', key)

    def get_frequency(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetFrequency', port_number, mode_number)

    def get_cutoff_frequency(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetFcutoff', port_number, mode_number)

    def get_mode_type(self, port_number: int, mode_number: int) -> str:
        return self.invoke_method('GetModeType', port_number, mode_number)

    def get_beta(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetBeta', port_number, mode_number)

    def get_alpha(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetAlpha', port_number, mode_number)

    def get_accuracy(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetAccuracy', port_number, mode_number)

    def get_wave_impedance(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetWaveImpedance', port_number, mode_number)

    def get_line_impedance(self, port_number: int, mode_number: int) -> float:
        return self.invoke_method('GetLineImpedance', port_number, mode_number)

    def get_line_impedance_broad_by_index(
            self, port_number: int, mode_number: int, index: int) -> float:
        return self.invoke_method('GetLineImpedanceBroadByIndex', port_number, mode_number, index)

    def get_line_impedance_broad_by_frequency(
            self, port_number: int, mode_number: int, frequency: float) -> float:
        return self.invoke_method(
            'GetLineImpedanceBroadByFreq', port_number, mode_number, frequency)

    def get_type(self, port_number: int) -> str:
        return self.invoke_method('GetType', port_number)

    def get_number_of_modes(self, port_number: int) -> int:
        return self.invoke_method('GetNumberOfModes', port_number)

    def get_port_mesh_location(self, port_number: int):
        orientation = w32com.create_ref_long()
        x_min = w32com.create_ref_long()
        y_min = w32com.create_ref_long()
        z_min = w32com.create_ref_long()
        x_max = w32com.create_ref_long()
        y_max = w32com.create_ref_long()
        z_max = w32com.create_ref_long()
        self.invoke_method(
            'GetPortMeshLocation', port_number, orientation,
            x_min, x_max, y_min, y_max, z_min, z_max)
        return (
            orientation.value,
            x_min.value, x_max.value, y_min.value,
            y_max.value, z_min.value, z_max.value)

    def get_port_mesh_coordinates(self, port_number: int):
        orientation = w32com.create_ref_long()
        x_min = w32com.create_ref_double()
        y_min = w32com.create_ref_double()
        z_min = w32com.create_ref_double()
        x_max = w32com.create_ref_double()
        y_max = w32com.create_ref_double()
        z_max = w32com.create_ref_double()
        self.invoke_method(
            'GetPortMeshCoordinates', port_number, orientation,
            x_min, x_max, y_min, y_max, z_min, z_max)
        return (
            orientation.value,
            x_min.value, x_max.value, y_min.value,
            y_max.value, z_min.value, z_max.value)

    def get_port_center_coordinates(self, port_number: int):
        x = w32com.create_ref_double()
        y = w32com.create_ref_double()
        z = w32com.create_ref_double()
        self.invoke_method('GetPortCenterCoordinates', port_number, x, y, z)
        return (x.value, y.value, z.value)

    def get_face_port_type_and_size(self, port_number: int):
        face_port_type = w32com.create_ref_long()
        size1 = w32com.create_ref_double()
        size2 = w32com.create_ref_double()
        self.invoke_method('GetFacePortTypeAndSize', port_number, face_port_type, size1, size2)
        return (face_port_type.value, size1.value, size2.value)

    def get_label(self, port_number: int):
        return self.invoke_method('GetLabel', port_number)


# from . import info
from . import \
    IProject, IVBAProvider, Modeler, Schematic, InterfacePCBS, HistoryListGenerator, Parameters, \
    VBATypeName, IQuietModeController
from enum import Enum
from typing import Optional, Union
import os.path

def cast_vba_value(value: str, type_name: str):
    if type_name in ['Byte', 'Integer', 'Long']:
        return int(value)
    elif type_name in ['String', 'Variant']:
        return value
    elif type_name in ['Single', 'Double']:
        return float(value)
    elif type_name == 'Boolean':
        return value != 'False'
    else:
        raise RuntimeError('Unsupported VBA Type Name')

def get_types_vba_t(
        return_type: Optional[VBATypeName], *args: Union[str, int, float, bool, VBATypeName]):
    res = []
    if return_type is not None:
        res.append(return_type.value)
    for arg in args:
        if isinstance(arg, VBATypeName):
            res.append(arg.value)
    return res

class Project(IProject, IVBAProvider):
    SOLVER_HF_TIME_DOMAIN = 'HF Time Domain'
    SOLVER_HF_EIGENMODE = 'HF Eigenmode'
    SOLVER_HF_FREQUENCY_DOMAIN = 'HF Frequency Domain'
    SOLVER_HF_INTEGRAL_EQ = 'HF IntegralEq'
    SOLVER_HF_MULTILAYER = 'HF Multilayer'
    SOLVER_HF_ASYMPTOTIC = 'HF Asymptotic'
    SOLVER_LF_E_STATIC = 'LF EStatic'
    SOLVER_LF_M_STATIC = 'LF MStatic'
    SOLVER_LF_STATIONARY_CURRENT = 'LF Stationary Current'
    SOLVER_LF_FREQUENCY_DOMAIN = 'LF Frequency Domain'
    SOLVER_LF_TIME_DOMAIN_MQS = 'LF Time Domain (MQS)'
    SOLVER_PT_TRACKING = 'PT Tracking'
    SOLVER_PT_WAKEFIELDS = 'PT Wakefields'
    SOLVER_PT_PIC = 'PT PIC'
    SOLVER_THERMAL_STEADY_STATE = 'Thermal Steady State'
    SOLVER_THERMAL_TRANSIENT = 'Thermal Transient'
    SOLVER_MECHANICS = 'Mechanics'

    GUIDE_TYPE_TRANSIENT = 'Transient'
    GUIDE_TYPE_EIGENMODE = 'Eigenmode'
    GUIDE_TYPE_FREQUENCY_DOMAIN = 'Frequency Domain'
    GUIDE_TYPE_ELECTROSTATIC = 'Electrostatic'
    GUIDE_TYPE_MAGNETOSTATIC = 'Magnetostatic'
    GUIDE_TYPE_STATIONARY_CURRENT = 'Stationary Current'
    GUIDE_TYPE_LOW_FREQUENCY = 'Low Frequency'
    GUIDE_TYPE_PARTICLE_TRACKING = 'Particle Tracking'

    APP_NAME_EMS = 'EMS'
    APP_NAME_PS = 'PS'
    APP_NAME_MWS = 'MWS'
    APP_NAME_MS = 'MS'
    APP_NAME_DS_FOR_MWS = 'DS for MWS'
    APP_NAME_DS_FOR_PCBS = 'DS for PCBS'
    APP_NAME_DS_FOR_CS = 'DS for CS'
    APP_NAME_DS_FOR_MS = 'DS for MS'
    APP_NAME_DS = 'DS'

    FILTER_NAME_RESULT_NAME = 'resultname'
    FILTER_NAME_TYPE = 'type'
    FILTER_NAME_TEMPLATE_NAME = 'templatename'
    FILTER_NAME_FOLDER = 'folder'

    VAL_TYPE_0D = '0D'
    VAL_TYPE_1D = '1D'
    VAL_TYPE_1D_COMPLEX = '1DC'
    VAL_TYPE_M0D = 'M0D'
    VAL_TYPE_M1D = 'M1D'
    VAL_TYPE_M1D_COMPLEX = 'M1DC'

    ARCHITECTURE_WIN32 = 'Windows IA32'
    ARCHITECTURE_WIN64 = 'Windows AMD64'
    ARCHITECTURE_LINUX32 = 'Linux IA32'
    ARCHITECTURE_LINUX64 = 'Linux AMD64'

    PATH_TYPE_ROOT = 'Root'
    PATH_TYPE_PROJECT = 'Project'
    PATH_TYPE_MODEL_3D = 'Model3D'
    PATH_TYPE_MODEL_CACHE = 'ModelCache'
    PATH_TYPE_RESULT = 'Result'
    PATH_TYPE_TEMP = 'Temp'

    class Kind(Enum):
        from cst.interface import ProjectType
        DesignStudio = ProjectType.DS
        MicrowaveStudio = ProjectType.MWS
        EMStudio = ProjectType.EMS
        ParticleStudio = ProjectType.PS
        MphysicsStudio = ProjectType.MPS
        CableStudio = ProjectType.CS
        PCBStudio = ProjectType.PCBS
        FilterDesigner3D = ProjectType.FD3D

    def __init__(self, native_obj, quiet_mode_controller: IQuietModeController) -> None:
        self.native_obj = native_obj
        self.quiet_mode_controller = quiet_mode_controller
        self.history_list_generator = HistoryListGenerator(self.get_modeler())
        # self.report_information(self.__get_welcome_message())

    def get_history_list_generator(self) -> HistoryListGenerator:
        return self.history_list_generator

    def run_vba(self, vba_code: str) -> bool:
        modeler = self.get_modeler()
        if modeler is None:
            return False
        else:
            return modeler.run_vba(vba_code)

    def invoke_function(self, function_name: str, *args: Union[str, int, float, bool]) -> bool:
        modeler = self.get_modeler()
        if modeler is None:
            return False
        else:
            return modeler.invoke_function(function_name, *args)

    def invoke_method(
            self, object_name: str, function_name: str, *args: Union[str, int, float, bool]):
        modeler = self.get_modeler()
        if modeler is None:
            return False
        else:
            return modeler.invoke_method(object_name, function_name, *args)

    def query_vba(self, vba_code: str) -> Optional[str]:
        modeler = self.get_modeler()
        if modeler is None:
            return None
        else:
            modeler.report_vba(vba_code)
            return self.get_last_message_text()

    def query_function(
            self, function_name: str, *args: Union[str, int, float, bool]) -> Optional[str]:
        modeler = self.get_modeler()
        if modeler is None:
            return None
        else:
            modeler.report_function(function_name, *args)
            return self.get_last_message_text()

    def query_method(
            self, object_name: str, function_name: str,
            *args: Union[str, int, float, bool]) -> Optional[str]:
        modeler = self.get_modeler()
        if modeler is None:
            return None
        else:
            modeler.report_method(object_name, function_name, *args)
            return self.get_last_message_text()

    def query_function_t(
            self, function_name: str, return_type: Optional[VBATypeName],
            *args: Union[str, int, float, bool, VBATypeName]) -> tuple:
        modeler = self.get_modeler()
        if modeler is None:
            return None
        else:
            modeler.report_function_t(function_name, return_type, *args)
            type_names = get_types_vba_t(return_type, *args)
            res_parts = self.get_last_message_text().split(',')
            if len(type_names) != len(res_parts):
                raise RuntimeError('Invalid number of return values')
            res = []
            for i in range(len(type_names)):
                res.append(cast_vba_value(res_parts[i], type_names[i]))
            return tuple(res)

    def query_method_t(
            self, object_name: str, function_name: str, return_type: Optional[VBATypeName],
            *args: Union[str, int, float, bool, VBATypeName]) -> tuple:
        modeler = self.get_modeler()
        if modeler is None:
            return None
        else:
            modeler.report_method_t(object_name, function_name, return_type, *args)
            type_names = get_types_vba_t(return_type, *args)
            res_parts = self.get_last_message_text().split(',')
            if len(type_names) != len(res_parts):
                raise RuntimeError('Invalid number of return values')
            res = []
            for i in range(len(type_names)):
                res.append(cast_vba_value(res_parts[i], type_names[i]))
            return tuple(res)

    def get_quiet_mode_controller(self) -> IQuietModeController:
        return self.quiet_mode_controller

    def activate(self) -> None:
        self.native_obj.activate()

    def close(self) -> None:
        self.native_obj.close()

    def get_file_name(self) -> str:
        return self.native_obj.filename()

    def get_folder(self) -> str:
        return self.native_obj.folder()

    def get_messages(self) -> list:
        msgs = self.native_obj.get_messages()
        return [] if msgs is None else msgs

    def get_last_message_text(self) -> Optional[str]:
        msgs = self.get_messages()
        return None if len(msgs) == 0 else msgs[-1]['text']

    def get_kind(self) -> Kind:
        return Project.Kind(self.native_obj.project_type())

    def save(self, path: str, include_results: bool = True) -> None:
        self.native_obj.save(path, include_results)

    def get_modeler(self) -> Modeler:
        return None if self.native_obj.modeler is None else Modeler(self.native_obj.modeler)

    def get_schematic(self) -> Schematic:
        return None if self.native_obj.schematic is None else Schematic(self.native_obj.schematic)

    def get_interface_pcbs(self) -> InterfacePCBS:
        return None if self.native_obj.pcbs is None else InterfacePCBS(self.native_obj.pcbs)

    def generate_history_list_from_cache(self) -> None:
        self.get_history_list_generator().generate_cache()

    def get_parameters(self) -> Parameters:
        return Parameters(self)

    def reset_all(self):
        self.invoke_function('ResetAll')

    def set_screen_updating(self, flag: bool = True):
        self.invoke_function('ScreenUpdating', flag)

    def select_quick_start_guide(self, guide_type: str):
        self.invoke_function('SelectQuickStartGuide', guide_type)

    def select_tree_item(self, item_name: str):
        self.quiet_mode_controller.store_and_disable_quiet_mode()
        self.invoke_function('SelectTreeItem', item_name)
        self.quiet_mode_controller.restore_quiet_mode()

    def get_number_of_selected_tree_items(self) -> int:
        return int(self.query_function('GetNumberOfSelectedTreeItems'))

    def get_selected_tree_item(self) -> str:
        return self.query_function('GetSelectedTreeItem')

    def get_next_selected_tree_item(self) -> str:
        return self.query_function('GetNextSelectedTreeItem')

    def set_lock(self, flag: bool = True):
        self.invoke_function('SetLock', flag)

    def lock(self, flag):
        self.invoke_function('SetLock', True)

    def unlock(self):
        self.invoke_function('SetLock', False)

    def save_as(self, new_path: str, include_results: bool = False):
        new_path = os.path.abspath(new_path)
        self.invoke_function('SaveAs', new_path, include_results)

    def store_in_archive(
            self, path: str, keep_all_results: bool = True, keep_1d_results: bool = True,
            keep_farfield_data: bool = True, delete_project_folder: bool = False):
        self.invoke_function('StoreInArchive',
            path, keep_all_results, keep_1d_results, keep_farfield_data, delete_project_folder)

    def clear_global_data_values(self):
        self.invoke_function('ClearGlobalDataValues')

    def delete_global_data_values(self, name: str):
        self.invoke_function('DeleteGlobalDataValue', name)

    def restore_global_data_value(self, name: str) -> str:
        return self.query_function('RestoreGlobalDataValue', name)

    def store_global_data_value(self, name: str, value):
        self.invoke_function('StoreGlobalDataValue', name, value)

    def reset_global_data_storage(self):
        self.invoke_function('ResetGlobalDataStorage')

    def set_global_data(self, name: str, value):
        self.invoke_function('SetGlobalData', name, value)

    def get_global_data(self, name: str) -> str:
        return self.query_function('GetGlobalData', name)

    def evaluate(self, expression) -> str:
        return self.query_function('Evaluate', expression)

    def set_activate_script_settings(self, switch: bool = True):
        self.invoke_function('ActivateScriptSettings', switch)

    def clear_script_settings(self):
        self.invoke_function('ClearScriptSettings')

    def get_script_setting(self, name: str, default_value: str) -> str:
        return self.query_function('GetScriptSetting', name, default_value)

    def store_script_setting(self, name: str, value: str) -> str:
        self.invoke_function('StoreScriptSetting', name, value)

    def store_template_setting(self, name: str, value: str) -> str:
        self.invoke_function('StoreTemplateSetting', name, value)

    def get_script_file_name(self) -> str:
        return self.query_function('GetScriptFileName')

    def evaluate_result_templates(self):
        self.invoke_function('EvaluateResultTemplates')

    def set_application_name(self, name: str):
        self.invoke_function('SetApplicationName', name)

    def reset_application_name(self):
        self.invoke_function('ResetApplicationName')

    def reset_template_iterator(self):
        self.invoke_function('ResetTemplateIterator')

    def set_template_filter(self, filter_name: str, value: str):
        self.invoke_function('SetTemplateFilter', filter_name, value)

    def get_next_template(
            self, result_name: str, val_type: str, template_name: str, folder: str) -> bool:
        return self.query_function(
            'GetNextTemplate', result_name, val_type, template_name, folder) != 'False'

    def get_file_type(self, file_path: str) -> str:
        return self.query_function('GetFileType', file_path)

    def get_first_table_result(self, result_name: str) -> str:
        return self.query_function('GetFirstTableResult', result_name)

    def get_next_table_result(self, result_name: str) -> str:
        return self.query_function('GetNextTableResult', result_name)

    def was_template_aborted(self) -> bool:
        return self.query_function('GetTemplateAborted') != 'False'

    def get_macro_path(self) -> str:
        return self.invoke_function('GetMacroPath')

    def get_macro_path_from_index(self, index: int) -> str:
        return self.invoke_function('GetMacroPathFromIndex', index)

    def get_number_of_macro_paths(self) -> int:
        return int(self.invoke_function('GetNumberOfMacroPaths'))

    def run_and_wait(self, command: str):
        self.invoke_function('RunAndWait', command)

    def run_macro(self, macro_name: str):
        self.invoke_function('RunMacro', macro_name)

    def run_script(self, script_name: str):
        self.invoke_function('RunScript', script_name)

    def report_information(self, *msg):
        self.invoke_function('ReportInformation', ''.join(list(map(lambda x: str(x), msg))))

    def report_warning(self, *msg):
        self.invoke_function('ReportWarning', ''.join(list(map(lambda x: str(x), msg))))

    def report_information_to_window(self, *msg):
        self.invoke_function('ReportInformationToWindow', ''.join(list(map(lambda x: str(x), msg))))

    def report_warning_to_window(self, *msg):
        self.invoke_function('ReportWarningToWindow', ''.join(list(map(lambda x: str(x), msg))))

    def report_error(self, *msg):
        self.invoke_function('ReportError', ''.join(list(map(lambda x: str(x), msg))))

    def set_common_mpi_cluster_config(
            self, install_folder: str, temp_folder: str, architecture: str):
        self.invoke_function('SetCommonMPIClusterConfig', install_folder, temp_folder, architecture)

    def set_common_mpi_cluster_login_info(self, user_name: str, private_key_file_path: str):
        self.invoke_function('SetCommonMPIClusterLoginInfo', user_name, private_key_file_path)

    def get_mpi_cluster_size(self) -> int:
        return int(self.query_function('GetMPIClusterSize'))

    def clear_mpi_cluster_config(self):
        self.invoke_function('ClearMPIClusterConfig')

    def add_mpi_cluster_node_config(
            self, host_name: str, install_folder: str, temp_folder: str, architecture: str,
            active: bool = True):
        self.invoke_function(
            'AddMPIClusterNodeConfig', host_name, install_folder, temp_folder,
            architecture, active)

    def import_xy_curve_from_ascii_file(self, folder_name: str, file_path: str):
        self.invoke_function('ImportXYCurveFromASCIIFile', folder_name, file_path)

    def paste_curves_from_ascii_file(self, folder_name: str, file_path: str):
        self.invoke_function('PasteCurvesFromASCIIFile', folder_name, file_path)

    def paste_curves_from_clipboard(self, folder_name: str):
        self.invoke_function('PasteCurvesFromClipboard', folder_name)

    def scale_curves(self, folder_name: str, x_scale: float, y_scale: float):
        self.invoke_function('ScaleCurves', folder_name, x_scale, y_scale)

    def store_curves_in_ascii_files(self, file_name: str):
        self.invoke_function('StoreCurvesInASCIIFile', file_name)

    def store_curves_in_clipboard(self):
        self.invoke_function('StoreCurvesInClipboard')

    def delete_results(self):
        self.invoke_function('DeleteResults')

    def get_field_frequency(self) -> float:
        return float(self.query_function('GetFieldFrequency'))

    def get_application_name(self) -> str:
        return self.query_function('GetApplicationName')

    def get_application_version(self) -> str:
        return self.query_function('GetApplicationVersion')

    def get_install_path(self) -> str:
        return self.query_function('GetInstallPath')

    def get_project_path(self, path_type: str) -> str:
        return self.query_function('GetProjectPath', path_type)

    def is_building_model(self) -> bool:
        return self.query_function('IsBuildingModel') != 'False'

    def get_license_host_id(self) -> str:
        return self.query_function('GetLicenseHostId')

    def get_license_customer_number(self) -> str:
        return self.query_function('GetLicenseCustomerNumber')

    def rebuild(self):
        self.invoke_function('Rebuild')

    def rebuild_on_parametric_change(self):
        show_msg_box = self.interface.is_in_quiet_mode()
        full_rebuild = False
        self.invoke_function('RebuildOnParametricChange', full_rebuild, show_msg_box)

    def rebuild_fully_on_parametric_change(self):
        show_msg_box = self.interface.is_in_quiet_mode()
        full_rebuild = True
        self.invoke_function('RebuildOnParametricChange', full_rebuild, show_msg_box)

    def get_pickpoint_coords(self, pickpoint_id: int) -> tuple[float, float, float]:
        x = float(self.query_function('xp', pickpoint_id))
        y = float(self.query_function('yp', pickpoint_id))
        z = float(self.query_function('zp', pickpoint_id))
        return (x, y, z)

    def export_image_to_file(self, path: str, width: int = 0, height: int = 0):
        self.invoke_function('ExportImageToFile', path, width, height)

    def export_image_to_clipboard(self, width: int = 0, height: int = 0):
        self.invoke_function('ExportImageToClipboard', width, height)

    # def __get_welcome_message(self):
    #     return \
    #         f'CST Studio Suite Python interface v{info.MODULE_VERSION} by '+\
    #         f'Samuel Travnicek connected.'


from . import IHistoryListProvider, IVBARuntime

class Modeler(IHistoryListProvider, IVBARuntime):
    def __init__(self, native_obj) -> None:
        self.native_obj = native_obj

    def abort_solver(self) -> None:
        self.native_obj.abort_solver()

    def run_vba(self, vba_code: str) -> bool:
        return self.native_obj._execute_vba_code(f'sub main\n{vba_code}\nend sub')

    def report_vba(self, vba_code: str) -> bool:
        return self.native_obj\
            ._execute_vba_code(f'sub main\nReportInformation({vba_code} & "")\nend sub')

    def add_to_history(self, header: str, vba_code: str) -> None:
        self.native_obj.add_to_history(header, vba_code)

    def full_history_rebuild(self) -> None:
        self.native_obj.full_history_rebuild(self)

    def get_active_solver_name(self) -> str:
        return self.native_obj.get_active_solver_name()

    def get_solver_run_info(self) -> dict:
        return self.native_obj.get_solver_run_info()

    def get_tree_items(self) -> list:
        return self.native_obj.get_tree_items()

    def is_solver_running(self) -> bool:
        return self.native_obj.is_solver_running()

    def pause_solver(self) -> None:
        self.native_obj.pause_solver()

    def resume_solver(self) -> None:
        self.native_obj.resume_solver()

    # blocking
    def run_solver(self) -> None:
        self.native_obj.run_solver()

    # non-blocking, use in combination with is_solver_running()
    def start_solver(self) -> None:
        self.native_obj.start_solver()


from .CST import IVBAProvider, Mesh, MeshSettings

def set_hexahedral_type(vbap: IVBAProvider):
    mesh = Mesh(vbap)
    mesh.set_mesh_type(Mesh.TYPE_PBA)
    mesh_settings = MeshSettings(vbap)
    mesh_settings.set_mesh_type(MeshSettings.MESH_TYPE_HEX)



from . import Project
from . import ComObjectWrapper

class TraceFromCurve(ComObjectWrapper):
    GAP_TYPE_ROUNDED = 0
    GAP_TYPE_EXTENDED = 1
    GAP_TYPE_NATURAL = 2

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.TraceFromCurve

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_component(self, component_name: str):
        self.invoke_method('Component', component_name)

    def set_material(self, mat_name: str):
        self.invoke_method('Material', mat_name)

    def set_curve(self, curve_name: str):
        self.invoke_method('Curve', curve_name)

    def set_thickness(self, thickness: float):
        self.invoke_method('Thickness', thickness)

    def set_width(self, width: float):
        self.invoke_method('Width', width)

    def set_start_round(self, flag: bool = True):
        self.invoke_method('RoundStart', flag)

    def set_end_round(self, flag: bool = True):
        self.invoke_method('RoundEnd', flag)

    def set_gap_type(self, gap_type: int):
        self.invoke_method('GapType', gap_type)

    def create(self):
        #TODO: check if Create or CreateNew
        self.invoke_method('Create')


__ALL__ = [

]

from .s_parameters import SParameters
from .s_parameters import SParameter
from .farfields import Farfields
from .farfields import Farfield
from .farfield_exporter import FarfieldExporter
from .uv_sphere_farfield_exporter import UVSphereFarfieldExporter
from .ascii_farfield_exporter import ASCIIFarfieldExporter


from . import IVBAProvider, VBAObjWrapper
import numpy as np

class Extrude(VBAObjWrapper):
    MODE_POINT_LIST = 'pointlist'
    MODE_PICKS = 'picks'
    MODE_MULTIPLE_PICKS = 'multiplepicks'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Extrude')

    def reset(self):
        self.record_method('Reset')

    def set_name(self, name: str):
        self.record_method('Name', name)

    def set_component(self, component_name: str):
        self.record_method('Component', component_name)

    def set_material(self, mat_name: str):
        self.record_method('Material', mat_name)

    def set_mode(self, mode: str):
        self.record_method('Mode', mode)

    def set_height(self, height: float):
        self.record_method('Height', height)

    def set_origin(self, x: float, y: float, z: float):
        self.record_method('Origin', x, y, z)

    def set_u_vector(self, u: float, v: float, w: float):
        self.record_method('Uvector', u, v, w)

    def set_v_vector(self, u: float, v: float, w: float):
        self.record_method('Vvector', u, v, w)

    def set_first_point(self, u: float, v: float):
        self.record_method('Point', u, v)

    def add_line_to(self, u: float, v: float):
        self.record_method('LineTo', u, v)

    def add_line_relative(self, u: float, v: float):
        self.record_method('RLine', u, v)

    def set_taper(self, taper: float):
        self.record_method('Taper', taper)

    def set_twist_deg(self, angle_deg: float):
        self.record_method('Twist', angle_deg)

    def set_twist_rad(self, angle_rad: float):
        self.set_twist_deg(np.rad2deg(angle_rad))

    def set_use_picks_for_height(self, flag: bool = True):
        self.record_method('UsePicksForHeight', flag)

    def set_pick_height_determined_by_first_face(self, flag: bool = True):
        self.record_method('PickHeightDeterminedByFirstFace', flag)

    def set_num_picked_faces(self, count: int):
        self.record_method('NumberOfPickedFaces', count)

    def set_delete_base_face_solid(self, flag: bool = True):
        self.record_method('DeleteBaseFaceSolid', flag)

    def set_clear_picked_face(self, flag: bool = True):
        self.record_method('ClearPickedFace', flag)

    def modify_height(self):
        self.record_method('ModifyHeight')

    def set_simplify_active(self, flag: bool = True):
        self.record_method('SetSimplifyActive', flag)

    def set_simplify_min_points_arc(self, count: int):
        self.record_method('SetSimplifyMinPointsArc', count)

    def set_simplify_min_points_circle(self, count: int):
        self.record_method('SetSimplifyMinPointsCircle', count)

    def set_simplify_angle_deg(self, angle_deg: float):
        self.record_method('SetSimplifyAngle', angle_deg)

    def set_simplify_angle_rad(self, angle_rad: float):
        self.set_simplify_angle_deg(np.rad2deg(angle_rad))

    def set_simplify_adjacent_tol_deg(self, angle_deg: float):
        self.record_method('SetSimplifyAdjacentTol', angle_deg)

    def set_simplify_adjacent_tol_rad(self, angle_rad: float):
        self.set_simplify_adjacent_tol_deg(np.rad2deg(angle_rad))

    def set_simplify_radius_tol(self, deviation: float):
        self.record_method('SetSimplifyRadiusTol', deviation)

    def set_simplify_angle_tang_deg(self, angle_deg: float):
        self.record_method('SetSimplifyAngleTang', angle_deg)

    def set_simplify_angle_tang_rad(self, angle_rad: float):
        self.set_simplify_angle_tang_deg(np.rad2deg(angle_rad))

    def set_simplify_edge_length(self, length: float):
        self.record_method('SetSimplifyEdgeLength', length)

    def create(self):
        self.record_method('Create')



from . import Project
from . import ComObjectWrapper
import numpy as np

class Solid(ComObjectWrapper):
    KEY_INSIDE = 'Inside'
    KEY_OUTSIDE = 'Outside'
    KEY_CENTERED = 'Centered'

    MESH_APPROXIMATION_PBA = 'PBA'
    MESH_APPROXIMATION_STAIRCASE = 'Staircase'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Solid

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def delete(self, solid_name: str):
        self.invoke_method('Delete', solid_name)

    def rename(self, old_name: str, new_name: str):
        self.invoke_method('Rename', old_name, new_name)

    def change_component(self, solid_name: str, component_name: str):
        self.invoke_method('ChangeComponent', solid_name, component_name)

    def change_material(self, solid_name: str, material_name: str):
        self.invoke_method('ChangeMaterial', solid_name, material_name)

    def set_use_individual_color(self, flag: bool = True):
        self.invoke_method('SetUseIndividualColor', flag)

    def change_individual_color_rgb(self, solid_name: str, r: float, g: float, b: float):
        self.invoke_method('ChangeIndividualColor', solid_name, r, g, b)

    def set_fast_model_update(self, flag: bool = True):
        self.invoke_method('FastModelUpdate', flag)

    def perform_add_operation(self, solid1_name: str, solid2_name: str):
        self.invoke_method('Add', solid1_name, solid2_name)

    def perform_insert_operation(self, solid1_name: str, solid2_name: str):
        # insert solid2 into solid1
        self.invoke_method('Insert', solid1_name, solid2_name)

    def perform_intersect_operation(self, solid1_name: str, solid2_name: str):
        self.invoke_method('Intersect', solid1_name, solid2_name)

    def perform_subtract_operation(self, solid1_name: str, solid2_name: str):
        # subtract solid2 from solid1
        self.invoke_method('Subtract', solid1_name, solid2_name)

    def merge_materials_of_component(self, component_or_solid_name: str):
        self.invoke_method('MergeMaterialsOfComponent', component_or_solid_name)

    def set_shape_visualization_accuracy2(self, acc: int):
        self.invoke_method('ShapeVisualizationAccuracy2', acc)

    def set_shape_visualization_offset(self, offset: int):
        self.invoke_method('ShapeVisualizationOffset', offset)

    def attach_active_wcs(self, solid_name: str):
        self.invoke_method('AttachActiveWCS', solid_name)

    def blend_edge(self, radius: float):
        self.invoke_method('BlendEdge', radius)

    def chamfer_edge1_deg(self, depth: float, angle_deg: float, face_id: int):
        self.invoke_method('ChamferEdge', depth, angle_deg, False, face_id)

    def chamger_edge1_rad(self, depth: float, angle_rad: float, face_id: int):
        self.chamfer_edge1_deg(depth, np.rad2deg(angle_rad), face_id)

    def chamfer_edge2_deg(self, depth: float, angle_deg: float, face_id: int):
        self.invoke_method('ChamferEdge', depth, angle_deg, True, face_id)

    def chamger_edge2_rad(self, depth: float, angle_rad: float, face_id: int):
        self.chamfer_edge2_deg(depth, np.rad2deg(angle_rad), face_id)

    def slice_shape(self, solid_name: str, component_name: str):
        self.invoke_method('SliceShape', solid_name, component_name)

    def split_shape(self, solid_name: str, component_name: str):
        self.invoke_method('SplitShape', solid_name, component_name)

    def thicken_sheet_advanced(
            self, solid_name: str, key: str, thickness: float, clear_picks: bool = False):
        self.invoke_method('ThickenSheetAdvanced', solid_name, key, thickness, clear_picks)

    def shell_advanced(
            self, solid_name: str, key: str, thickness: float, clear_picks: bool = False):
        self.invoke_method('ShellAdvanced', solid_name, key, thickness, clear_picks)

    def fill_up_space_advanced(self, solid_name: str, component_name: str, material_name: str):
        self.invoke_method('FillupSpaceAdvanced', solid_name, component_name, material_name)

    def move_selected_face(self, dx: float, dy: float, dz: float):
        self.invoke_method('MoveSelectedFace', dx, dy, dz)

    def move_selected_planar_face(self, offset: float):
        self.invoke_method('MoveSelectedPlanarFace', offset)

    def offset_selected_faces(self, offset: float):
        self.invoke_method('OffsetSelectedFaces', offset)

    def remove_selected_faces(self):
        self.invoke_method('RemoveSelectedFaces')

    def set_selected_face_radius(self, radius: float):
        self.invoke_method('SelectedFaceRadius', radius)

    def set_mesh_step_width(self, solid_name: str, dx: float, dy: float, dz: float):
        self.invoke_method('SetMeshStepWidth', solid_name, dx, dy, dz)

    def set_mesh_extend_width(self, solid_name: str, dx: float, dy: float, dz: float):
        self.invoke_method('SetMeshExtendwidth', solid_name, dx, dy, dz)

    def set_automesh_fixpoints(self, solid_name: str, flag: bool = True):
        self.invoke_method('SetAutomeshFixpoints', solid_name, flag)

    def set_material_based_refinement(self, solid_name: str, flag: bool = True):
        self.invoke_method('SetMaterialBasedRefinement', solid_name, flag)

    def set_mesh_approximation(self, solid_name: str, approx: str, default_type: bool = False):
        self.invoke_method('SetMeshProperties', solid_name, approx, default_type)

    def set_use_for_simulation(self, solid_name: str, flag: bool = True):
        self.invoke_method('SetUseForSimulation', solid_name, flag)

    def set_use_thin_sheet_mesh_for_shape(self, solid_name: str, flag: bool = True):
        self.invoke_method('SetUseThinSheetMeshForShape', solid_name, flag)

    def set_mesh_refinement(
            self, solid_name: str, edge_refinement_factor: float, volume_refinement_factor: float):
        edge_refinement = edge_refinement_factor is not None
        volume_refinement = volume_refinement_factor is not None
        if not edge_refinement:
            edge_refinement_factor = 0
        if not volume_refinement:
            volume_refinement_factor = 0
        self.invoke_method(
            'SetMeshRefinement', edge_refinement, edge_refinement_factor,
            volume_refinement, volume_refinement_factor)

    # def set_solid_local_mesh_properties(self, )